import os
import time
import hashlib
import base64

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.defer import Deferred  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust_forks.Bismuth.bismuthclient import bismuthclient  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import jsn

from bitdust.main import settings
from bitdust.main import events

from bitdust.system import local_fs

from bitdust.blockchain import known_bismuth_nodes

from bitdust.services import driver

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

_BismuthClient = None
_DataDirPath = None
_MyTransactionsDirPath = None
_MyLatestTransactionBlockHeight = 0
_MyKnownTransactions = []

#------------------------------------------------------------------------------


def init():
    global _BismuthClient
    global _DataDirPath
    global _MyTransactionsDirPath
    _DataDirPath = settings.ServiceDir('bismuth_blockchain')
    _MyTransactionsDirPath = os.path.join(_DataDirPath, 'transactions')
    if not os.path.exists(_MyTransactionsDirPath):
        os.makedirs(_MyTransactionsDirPath)
    if driver.is_enabled('service_bismuth_node'):
        servers_list = ['127.0.0.1:15658']
    else:
        servers_list = ['{}:{}'.format(k, v) for k, v in known_bismuth_nodes.nodes_by_host().items()]
    _BismuthClient = bismuthclient.BismuthClient(
        servers_list=servers_list,
        wallet_file=wallet_file_path(),
        verbose=(_Debug and (_DebugLevel >= 12)),
    )
    ret = Deferred()
    reactor.callLater(0, check_create_wallet, ret)  # @UndefinedVariable
    return ret


def shutdown():
    global _BismuthClient
    global _DataDirPath
    del _BismuthClient
    _BismuthClient = None
    _DataDirPath = None


#------------------------------------------------------------------------------


def client():
    global _BismuthClient
    return _BismuthClient


def data_dir():
    global _DataDirPath
    return _DataDirPath


def wallet_file_path(wallet_name=None):
    if not wallet_name:
        wallet_name = 'wallet'
    return os.path.join(data_dir(), wallet_name + '_key.json')


def my_transactions_dir():
    global _MyTransactionsDirPath
    return _MyTransactionsDirPath


def my_known_transactions():
    global _MyKnownTransactions
    return _MyKnownTransactions


#------------------------------------------------------------------------------


def check_create_wallet(result_defer):
    file_path = wallet_file_path()
    if _Debug:
        lg.args(_DebugLevel, file_path=file_path)
    if os.path.isfile(file_path):
        if _Debug:
            lg.dbg(_DebugLevel, 'wallet file already exists')
    else:
        if client().new_wallet(file_path):
            client().load_wallet(file_path)
        else:
            result_defer.errback(Exception('error creating wallet'))
            return

    success = None
    count = 0
    while True:
        if count > 10:
            success = False
            break
        try:
            cur_balance = client().balance()
        except Exception as e:
            lg.warn(e)
            time.sleep(5)
            count += 1
            continue
        if _Debug:
            lg.args(_DebugLevel, cur_balance=cur_balance, my_wallet_address=my_wallet_address())
        if cur_balance == 'N/A':
            time.sleep(5)
            count += 1
            continue
        success = True
        break

    if not success:
        result_defer.errback(Exception('error connecting to Bismuth node'))
        return
    result_defer.callback(True)


def my_wallet_address():
    return client().address


def my_balance():
    try:
        _balance = float(client().balance())
    except:
        lg.exc()
        return 'N/A'
    return _balance


def latest_transactions(num, offset, for_display, mempool_included):
    return client().latest_transactions(num, offset, for_display, mempool_included)


def send_transaction(recipient, amount, operation='', data='', raise_errors=False):
    error_reply = []
    ret = client().send(recipient=recipient, amount=amount, operation=operation, data=data, error_reply=error_reply)
    if not ret:
        if raise_errors:
            raise Exception(error_reply)
        return error_reply
    return ret


def find_transaction(address=None, recipient=None, operation=None, openfield=None, limit=10, offset=0, block_height_from=None):
    try:
        ret = client().search_transactions(
            address=address,
            recipient=recipient,
            operation=operation,
            openfield=openfield,
            limit=limit,
            offset=offset,
            block_height_from=block_height_from,
        )
    except:
        lg.exc()
        return []
    if _Debug:
        lg.args(_DebugLevel, a=address, r=recipient, o=operation, d=str(openfield)[:20], lim=limit, ofs=offset, ret=str(ret)[:20])
    return ret


def pack_transaction_id(tx):
    m = hashlib.blake2b(digest_size=4)
    m.update(tx['signature'].encode('utf-8'))
    return '{}_{}'.format(tx['block_height'], base64.b16encode(m.digest()).decode('utf-8').lower())


def sync_my_transactions():
    global _MyTransactionsDirPath
    global _MyLatestTransactionBlockHeight
    global _MyKnownTransactions
    all_known_transactions = set()
    for filename in os.listdir(_MyTransactionsDirPath):
        filepath = os.path.join(_MyTransactionsDirPath, filename)
        try:
            tx = jsn.loads_text(local_fs.ReadTextFile(filepath))
            int(tx['block_height'])
        except:
            lg.exc()
            continue
        all_known_transactions.add(filename)
        if _MyLatestTransactionBlockHeight < int(tx['block_height']):
            _MyLatestTransactionBlockHeight = int(tx['block_height'])
        _MyKnownTransactions.append((
            int(tx['block_height']),
            tx['operation'],
            tx['openfield'],
            filename,
        ))
        if len(_MyKnownTransactions) > 500:
            _MyKnownTransactions.pop(0)
    blockchain_transactions = []
    blockchain_transactions.extend(find_transaction(address=my_wallet_address(), block_height_from=_MyLatestTransactionBlockHeight))
    blockchain_transactions.extend(find_transaction(recipient=my_wallet_address(), block_height_from=_MyLatestTransactionBlockHeight))
    new_transactions = 0
    for tx in blockchain_transactions:
        filename = pack_transaction_id(tx)
        if filename in all_known_transactions:
            continue
        all_known_transactions.add(filename)
        if _MyLatestTransactionBlockHeight < int(tx['block_height']):
            _MyLatestTransactionBlockHeight = int(tx['block_height'])
        _MyKnownTransactions.append((
            int(tx['block_height']),
            tx['operation'],
            tx['openfield'],
            filename,
        ))
        if len(_MyKnownTransactions) > 500:
            _MyKnownTransactions.pop(0)
        filepath = os.path.join(_MyTransactionsDirPath, filename)
        local_fs.WriteTextFile(filepath, jsn.dumps(tx))
        new_transactions += 1
        tx['filename'] = filename
        events.send('blockchain-transaction-received', data=tx)
    if _Debug:
        lg.args(_DebugLevel, new_transactions=new_transactions, known_transactions=len(all_known_transactions), latest_block=_MyLatestTransactionBlockHeight)



import socks, connections
import options

config = options.Get()
config.read()
version = config.version

s = socks.socksocket()

port = 5658
if 'testnet' in version:
    port = 2829
    print('tesnet mode')
elif 'regnet' in version:
    is_regnet = True
    print('Regtest mode')
    port = 3030

while True:
    try:
        s.connect(('127.0.0.1', port))

        print('Sending stop command...')
        connections.send(s, 'stop')
        print('Stop command delivered.')
        break
    except:
        print('Cannot reach node, retrying...')

s.close()



#!/usr/bin/python
# message.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (message.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: message

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import time
import base64

#------------------------------------------------------------------------------

from twisted.internet.defer import fail
from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.main import config
from bitdust.main import settings

from bitdust.p2p import commands
from bitdust.p2p import online_status
from bitdust.p2p import p2p_service

from bitdust.lib import packetid
from bitdust.lib import utime
from bitdust.lib import serialization
from bitdust.lib import jsn
from bitdust.lib import strng

from bitdust.crypt import key
from bitdust.crypt import my_keys

from bitdust.contacts import identitycache

from bitdust.userid import id_url
from bitdust.userid import my_id
from bitdust.userid import global_id

#------------------------------------------------------------------------------

MAX_PENDING_MESSAGES_PER_CONSUMER = 100

#------------------------------------------------------------------------------

_ConsumersCallbacks = {}
_ReceivedMessagesIDs = []

_IncomingMessageCallbacks = []
_OutgoingMessageCallbacks = []

_MessageQueuePerConsumer = {}

_LastUserPingTime = {}
_PingTrustIntervalSeconds = 60*5

#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'message.init')
    AddIncomingMessageCallback(push_incoming_message)
    AddOutgoingMessageCallback(push_outgoing_message)


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'message.shutdown')
    RemoveOutgoingMessageCallback(push_outgoing_message)
    RemoveIncomingMessageCallback(push_incoming_message)


#------------------------------------------------------------------------------


def received_messages_ids(erase_old_records=False):
    global _ReceivedMessagesIDs
    if erase_old_records:
        _ReceivedMessagesIDs = _ReceivedMessagesIDs[50:]
    return _ReceivedMessagesIDs


def message_queue():
    global _MessageQueuePerConsumer
    return _MessageQueuePerConsumer


def consumers_callbacks():
    global _ConsumersCallbacks
    return _ConsumersCallbacks


#------------------------------------------------------------------------------


def ConnectCorrespondent(idurl):
    pass


def UniqueID():
    return str(int(time.time()*100.0))


#------------------------------------------------------------------------------


def AddIncomingMessageCallback(cb):
    """
    Calling with: (packet_in_object, private_message_object, decrypted_message_body)
    """
    global _IncomingMessageCallbacks
    if cb not in _IncomingMessageCallbacks:
        _IncomingMessageCallbacks.append(cb)
    else:
        lg.warn('callback method already exist')


def InsertIncomingMessageCallback(cb):
    """
    Calling with: (packet_in_object, private_message_object, decrypted_message_body)
    """
    global _IncomingMessageCallbacks
    if cb not in _IncomingMessageCallbacks:
        _IncomingMessageCallbacks.insert(0, cb)
    else:
        lg.warn('callback method already exist')


def RemoveIncomingMessageCallback(cb):
    global _IncomingMessageCallbacks
    if cb in _IncomingMessageCallbacks:
        _IncomingMessageCallbacks.remove(cb)
    else:
        lg.warn('callback method not exist')


def AddOutgoingMessageCallback(cb):
    """
    Calling with: (message_body, private_message_object, remote_identity, outpacket, packet_out_object)
    """
    global _OutgoingMessageCallbacks
    if cb not in _OutgoingMessageCallbacks:
        _OutgoingMessageCallbacks.append(cb)
    else:
        lg.warn('callback method already exist')


def RemoveOutgoingMessageCallback(cb):
    global _OutgoingMessageCallbacks
    if cb in _OutgoingMessageCallbacks:
        _OutgoingMessageCallbacks.remove(cb)
    else:
        lg.warn('callback method not exist')


#------------------------------------------------------------------------------


class PrivateMessage(object):

    """
    A class to represent a message.

    We always encrypt messages with a session key so we need to package
    with encrypted body.
    """

    def __init__(self, recipient, sender=None, encrypted_session=None, encrypted_body=None):
        self.sender = strng.to_text(sender or my_id.getGlobalID(key_alias='master'))
        self.recipient = strng.to_text(recipient)
        self.encrypted_session = encrypted_session
        self.encrypted_body = encrypted_body

    def __str__(self):
        return 'PrivateMessage(%s->%s)' % (
            self.sender,
            self.recipient,
        )

    def sender_id(self):
        return self.sender

    def recipient_id(self):
        return self.recipient

    def session_key(self):
        return self.encrypted_session

    def body(self):
        return self.encrypted_body

    def encrypt(self, message_body, encrypt_session_func=None):
        if _Debug:
            lg.args(_DebugLevel, encrypt_session_func=encrypt_session_func, recipient=self.recipient)
        new_sessionkey = key.NewSessionKey(session_key_type=key.SessionKeyType())
        if not encrypt_session_func:
            if my_keys.is_key_registered(self.recipient):
                if _Debug:
                    lg.dbg(_DebugLevel, 'with registered key %r' % self.recipient)
                encrypt_session_func = lambda inp: my_keys.encrypt(self.recipient, inp)
        if not encrypt_session_func:
            glob_id = global_id.NormalizeGlobalID(self.recipient)
            if glob_id['key_alias'] == 'master':
                if glob_id['idurl'] == my_id.getIDURL():
                    lg.warn('making encrypted message addressed to me ?')
                    encrypt_session_func = lambda inp: my_keys.encrypt('master', inp)
                else:
                    remote_identity = identitycache.FromCache(glob_id['idurl'])
                    if not remote_identity:
                        raise Exception('remote identity is not cached yet, not able to encrypt the message')
                    if _Debug:
                        lg.dbg(_DebugLevel, 'with remote identity public key %r' % glob_id['idurl'])
                    encrypt_session_func = remote_identity.encrypt
            else:
                own_key = global_id.MakeGlobalID(idurl=my_id.getIDURL(), key_alias=glob_id['key_alias'])
                if my_keys.is_key_registered(own_key):
                    if _Debug:
                        lg.dbg(_DebugLevel, 'with registered key (found by alias) %r' % own_key)
                    encrypt_session_func = lambda inp: my_keys.encrypt(own_key, inp)
        if not encrypt_session_func:
            raise Exception('can not find key for given recipient')
        self.encrypted_session = encrypt_session_func(new_sessionkey)
        self.encrypted_body = key.EncryptWithSessionKey(new_sessionkey, message_body, session_key_type=key.SessionKeyType())
        return self.encrypted_session, self.encrypted_body

    def decrypt(self, decrypt_session_func=None):
        if _Debug:
            lg.args(_DebugLevel, decrypt_session_func=decrypt_session_func, recipient=self.recipient)
        if not decrypt_session_func:
            if my_keys.is_key_registered(self.recipient):
                if _Debug:
                    lg.dbg(_DebugLevel, 'decrypt with registered key %r' % self.recipient)
                decrypt_session_func = lambda inp: my_keys.decrypt(self.recipient, inp)
        if not decrypt_session_func:
            glob_id = global_id.NormalizeGlobalID(self.recipient)
            if glob_id['idurl'] == my_id.getIDURL():
                if glob_id['key_alias'] == 'master':
                    if _Debug:
                        lg.dbg(_DebugLevel, 'decrypt with my master key %r' % self.recipient)
                    decrypt_session_func = lambda inp: my_keys.decrypt('master', inp)
        if not decrypt_session_func:
            raise Exception('can not find key for given recipient: %s' % self.recipient)
        decrypted_sessionkey = decrypt_session_func(self.encrypted_session)
        return key.DecryptWithSessionKey(decrypted_sessionkey, self.encrypted_body, session_key_type=key.SessionKeyType())

    def serialize(self):
        dct = {
            'r': self.recipient,
            's': self.sender,
            'k': strng.to_text(base64.b64encode(strng.to_bin(self.encrypted_session))),
            'p': self.encrypted_body,
        }
        return serialization.DictToBytes(dct, encoding='utf-8')

    @classmethod
    def deserialize(cls, input_string):
        try:
            dct = serialization.BytesToDict(input_string, keys_to_text=True, encoding='utf-8')
            message_obj = cls(
                recipient=strng.to_text(dct['r']),
                sender=strng.to_text(dct['s']),
                encrypted_session=base64.b64decode(strng.to_bin(dct['k'])),
                encrypted_body=dct['p'],
            )
        except:
            lg.exc()
            return None
        return message_obj


#------------------------------------------------------------------------------


class GroupMessage(PrivateMessage):

    def __str__(self):
        return 'GroupMessage(%s->%s)' % (
            self.sender,
            self.recipient,
        )


#------------------------------------------------------------------------------


def on_incoming_message(request, info, status, error_message):
    """
    Message came in for us
    """
    global _IncomingMessageCallbacks
    if _Debug:
        lg.out(_DebugLevel, 'message.on_incoming_message new PrivateMessage %r from %s' % (request.PacketID, request.OwnerID))
    private_message_object = PrivateMessage.deserialize(request.Payload)
    if private_message_object is None:
        lg.err('PrivateMessage deserialize failed, can not extract message from request payload of %d bytes' % len(request.Payload))
        return False
    try:
        decrypted_message = private_message_object.decrypt()
        json_message = serialization.BytesToDict(
            decrypted_message,
            unpack_types=True,
            encoding='utf-8',
        )
        json_message = jsn.dict_keys_to_text(jsn.dict_values_to_text(json_message))
    except Exception as exc:
        lg.err('decrypt %r failed: %r' % (private_message_object, exc))
        return False
    if request.PacketID in received_messages_ids():
        lg.warn('skip incoming message %s because found in recent history' % request.PacketID)
        return False
    received_messages_ids().append(request.PacketID)
    if len(received_messages_ids()) > 100:
        received_messages_ids(True)
    handled = False
    try:
        for cb in _IncomingMessageCallbacks:
            handled = cb(request, private_message_object, json_message)
            if _Debug:
                lg.args(_DebugLevel, cb=cb, packet_id=request.PacketID, handled=handled)
            if handled:
                break
    except:
        lg.exc()
    if _Debug:
        lg.args(_DebugLevel, msg_len=len(decrypted_message), handled=handled)
    if handled:
        return True
    if config.conf().getBool('services/private-messages/acknowledge-unread-messages-enabled'):
        p2p_service.SendAckNoRequest(request.OwnerID, request.PacketID, response='unread')
    return True


def on_ping_success(ok, idurl):
    global _LastUserPingTime
    idurl = id_url.to_bin(idurl)
    _LastUserPingTime[idurl] = time.time()
    lg.info('shake up hands %r before sending a message : %s' % (idurl, ok))
    return ok


def on_message_delivered(idurl, json_data, recipient_global_id, packet_id, response, info, result_defer=None):
    global _LastUserPingTime
    idurl = id_url.to_bin(idurl)
    if _Debug:
        lg.args(_DebugLevel, packet_id=packet_id, recipient_global_id=recipient_global_id)
    _LastUserPingTime[idurl] = time.time()
    if result_defer and not result_defer.called:
        result_defer.callback(response)


def on_message_failed(idurl, json_data, recipient_global_id, packet_id, response, info, result_defer=None, error=None):
    global _LastUserPingTime
    idurl = id_url.to_bin(idurl)
    lg.err('message %s failed sending to %s in %s because : %r' % (packet_id, recipient_global_id, response, error))
    if idurl in _LastUserPingTime:
        _LastUserPingTime[idurl] = 0
    if result_defer and not result_defer.called:
        err = Exception(response) if response else (error if not strng.is_string(error) else Exception(error))
        if _Debug:
            lg.args(_DebugLevel, err=err, i=idurl, r=recipient_global_id, pid=packet_id, j=json_data)
        result_defer.errback(err)
    return None


#------------------------------------------------------------------------------


def do_send_message(json_data, recipient_global_id, packet_id, message_ack_timeout, result_defer=None, fire_callbacks=True):
    global _OutgoingMessageCallbacks
    remote_idurl = global_id.GlobalUserToIDURL(recipient_global_id, as_field=False)
    if not remote_idurl:
        raise Exception('invalid recipient')
    remote_identity = identitycache.FromCache(remote_idurl)
    if not remote_identity:
        raise Exception('remote identity object not exist in cache')
    message_body = serialization.DictToBytes(
        json_data,
        pack_types=True,
        encoding='utf-8',
    )
    if _Debug:
        lg.out(_DebugLevel, 'message.do_send_message to %s with %d bytes message ack_timeout=%s pid:%s' % (recipient_global_id, len(message_body), message_ack_timeout, packet_id))
    try:
        private_message_object = PrivateMessage(recipient=recipient_global_id)
        private_message_object.encrypt(message_body)
    except:
        lg.exc()
        raise Exception('message encryption failed')
    payload = private_message_object.serialize()
    if _Debug:
        lg.out(_DebugLevel, '        payload is %d bytes, remote idurl is %s' % (len(payload), remote_idurl))
    callbacks = {}
    if message_ack_timeout:
        callbacks = {
            commands.Ack(): lambda response, info: on_message_delivered(
                remote_idurl,
                json_data,
                recipient_global_id,
                packet_id,
                response,
                info,
                result_defer,
            ),
            commands.Fail(): lambda response, info: on_message_failed(remote_idurl, json_data, recipient_global_id, packet_id, response, info, result_defer=result_defer, error='fail received'),
            None: lambda pkt_out: on_message_failed(
                remote_idurl,
                json_data,
                recipient_global_id,
                packet_id,
                None,
                None,
                result_defer=result_defer,
                error='timeout',
            ),
            'timeout': lambda pkt_out, errmsg: on_message_failed(
                remote_idurl,
                json_data,
                recipient_global_id,
                packet_id,
                None,
                None,
                result_defer=result_defer,
                error=errmsg,
            ),
            'failed': lambda pkt_out, errmsg: on_message_failed(
                remote_idurl,
                json_data,
                recipient_global_id,
                packet_id,
                None,
                None,
                result_defer=result_defer,
                error=errmsg,
            ),
        }
    result, outpacket = p2p_service.SendMessage(
        remote_idurl=remote_idurl,
        packet_id=packet_id,
        payload=payload,
        callbacks=callbacks,
        response_timeout=message_ack_timeout,
    )
    if fire_callbacks:
        try:
            for cp in _OutgoingMessageCallbacks:
                cp(json_data, private_message_object, remote_identity, outpacket, result)
        except:
            lg.exc()
            raise Exception('failed sending message')
    return result


def send_message(json_data, recipient_global_id, packet_id=None, message_ack_timeout=None, ping_timeout=None, ping_retries=0, skip_handshake=False, fire_callbacks=True, require_handshake=False):
    """
    Send command.Message() packet to remote peer. Returns Deferred object.
    """
    global _LastUserPingTime
    global _PingTrustIntervalSeconds
    if not packet_id:
        packet_id = packetid.UniqueID()
    if ping_timeout is None:
        ping_timeout = settings.P2PTimeOut()
    if message_ack_timeout is None:
        message_ack_timeout = settings.P2PTimeOut()
    if _Debug:
        lg.out(_DebugLevel, 'message.send_message to %s with PacketID=%s timeout=%d ack_timeout=%r retries=%d' % (recipient_global_id, packet_id, ping_timeout, message_ack_timeout, ping_retries))
    remote_idurl = global_id.GlobalUserToIDURL(recipient_global_id, as_field=False)
    if not remote_idurl:
        lg.warn('invalid recipient')
        return fail(Exception('invalid recipient'))
    ret = Deferred()
    ret.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='message.send_message')
    if remote_idurl not in _LastUserPingTime:
        is_ping_expired = True
    else:
        is_ping_expired = time.time() - _LastUserPingTime[remote_idurl] > _PingTrustIntervalSeconds
    remote_identity = identitycache.FromCache(remote_idurl)
    is_online = online_status.isOnline(remote_idurl)
    if _Debug:
        lg.out(_DebugLevel, '    is_ping_expired=%r  remote_identity=%r  is_online=%r  skip_handshake=%r' % (is_ping_expired, bool(remote_identity), is_online, skip_handshake))
    if require_handshake or remote_identity is None or ((is_ping_expired or not is_online) and not skip_handshake):
        d = online_status.handshake(
            idurl=remote_idurl,
            ack_timeout=ping_timeout,
            ping_retries=ping_retries,
            channel='send_message',
            keep_alive=True,
        )
        d.addCallback(lambda ok: on_ping_success(ok, remote_idurl))
        d.addCallback(lambda _: do_send_message(
            json_data=json_data,
            recipient_global_id=recipient_global_id,
            packet_id=packet_id,
            message_ack_timeout=message_ack_timeout,
            result_defer=ret,
            fire_callbacks=fire_callbacks,
        ))
        d.addErrback(lambda err: on_message_failed(remote_idurl, json_data, recipient_global_id, packet_id, None, None, result_defer=ret, error=err))
        return ret
    try:
        do_send_message(
            json_data=json_data,
            recipient_global_id=recipient_global_id,
            packet_id=packet_id,
            message_ack_timeout=message_ack_timeout,
            result_defer=ret,
            fire_callbacks=fire_callbacks,
        )
    except Exception as exc:
        lg.warn(str(exc))
        on_message_failed(remote_idurl, json_data, recipient_global_id, packet_id, None, None, error=exc)
        ret.errback(exc)
    return ret


#------------------------------------------------------------------------------


def consume_messages(consumer_callback_id, callback=None, direction=None, message_types=None, reset_callback=False):
    """
    Register a new callback method or Deferred object to wait and receive messages from the stream.

    If message was passed thru the stream but there were no callbacks registered to listen - the message is just ignored.
    When callback is registered any new message (if they match to specified criteria) will trigger callback to be executed.
    If callback is a Deferred object - the message will fire a callback() method of it so consumer can receive it.
    Right after receiving the message via callback() method consumer must call `consume_messages()` again to be able
    to receive the next message. This is very similar to a long polling technique.

    Stream also takes care of next messages that are not consumed yet by Deferred() callback in between of that calls to that method.
    If there are some messages that was not consumed by Deferred() - next call to that method will immediately fire
    callback() method of the Deferred() object with the list of messages.

    If input parameter `callback` is None Deferred() object will be automatically created and returned back as result.
    If input parameter `callback` is a callable method it will not be released like Deferred object - it will be fired
    for every message passed thru the stream.

    Parameter `direction` can be "incoming", "outgoing" or None (for both directions).

    Parameter `message_types` can be None (no filtering) or a list of desired types:
        "private_message", "group_message", "queue_message", "queue_message_replica"

    If `reset_callback` is True - previously registered Deferred object will be cleaned
    with `clear_consumer_callbacks()` method.
    """
    if consumer_callback_id in consumers_callbacks():
        if not reset_callback:
            raise Exception('consumer callback already exist')
        clear_consumer_callbacks(consumer_callback_id)
    cb = callback or Deferred()
    consumers_callbacks()[consumer_callback_id] = {
        'callback': cb,
        'direction': direction,
        'message_types': message_types,
    }
    if _Debug:
        lg.out(_DebugLevel, 'message.consume_messages added callback for consumer %r' % consumer_callback_id)
    do_read()
    return cb


def clear_consumer_callbacks(consumer_callback_id):
    if consumer_callback_id not in consumers_callbacks().keys():
        return True
    cb_info = consumers_callbacks().pop(consumer_callback_id)
    if isinstance(cb_info['callback'], Deferred):
        if _Debug:
            lg.args(_DebugLevel, consumer_callback_id=consumer_callback_id, cb=cb_info['callback'], called=cb_info['callback'].called)
        if not cb_info['callback'].called:
            cb_info['callback'].callback([])
            cb_info['callback'] = None
    else:
        if _Debug:
            lg.args(_DebugLevel, consumer_callback_id=consumer_callback_id, cb=cb_info['callback'])
    return True


#------------------------------------------------------------------------------


def push_message(direction, msg_type, recipient_id, sender_id, packet_id, owner_idurl, json_message, run_consumers=True):
    for consumers_callback_id in consumers_callbacks().keys():
        if consumers_callback_id not in message_queue():
            message_queue()[consumers_callback_id] = []
        message_queue()[consumers_callback_id].append(
            {
                'type': msg_type,
                'dir': direction,
                'to': recipient_id,
                'from': sender_id,
                'data': json_message,
                'packet_id': packet_id,
                'owner_idurl': owner_idurl,
                'time': utime.utcnow_to_sec1970(),
            }
        )
        if _Debug:
            lg.args(_DebugLevel, dir=direction, msg_type=msg_type, to_id=recipient_id, from_id=sender_id, cb=consumers_callback_id, pending=len(message_queue()[consumers_callback_id]))
    if not run_consumers:
        return 0
    total_consumed = do_read()
    return total_consumed > 0


#------------------------------------------------------------------------------


def push_incoming_message(request, private_message_object, json_message):
    msg_type = None
    if request.PacketID.startswith('private_'):
        msg_type = 'private_message'
    if request.PacketID.startswith('queue_'):
        msg_type = 'queue_message'
    if msg_type is None:
        raise Exception('undefined message type detected in %r' % request)
    return push_message(
        direction='incoming',
        msg_type=msg_type,
        recipient_id=private_message_object.recipient_id(),
        sender_id=private_message_object.sender_id(),
        packet_id=request.PacketID,
        owner_idurl=request.OwnerID,
        json_message=json_message,
    )


def push_outgoing_message(json_message, private_message_object, remote_identity, request, result):
    msg_type = 'private_message'
    if request.PacketID.startswith('queue_'):
        msg_type = 'queue_message'
    return push_message(
        direction='outgoing',
        msg_type=msg_type,
        recipient_id=private_message_object.recipient_id(),
        sender_id=private_message_object.sender_id(),
        packet_id=request.PacketID,
        owner_idurl=request.OwnerID,
        json_message=json_message,
    )


def push_group_message(json_message, direction, group_key_id, producer_id, sequence_id):
    if _Debug:
        lg.args(_DebugLevel, group_key_id=group_key_id, producer_id=producer_id, sequence_id=sequence_id)
    return push_message(
        direction=direction,
        msg_type='group_message',
        recipient_id=group_key_id,
        sender_id=producer_id,
        packet_id=sequence_id,
        owner_idurl=None,
        json_message=json_message,
    )


#------------------------------------------------------------------------------


def do_read():
    known_consumers = list(message_queue().keys())
    total_handled = 0
    for consumer_id in known_consumers:
        if consumer_id not in message_queue() or len(message_queue()[consumer_id]) == 0:
            continue
        cb_info = consumers_callbacks().get(consumer_id)
        pending_messages = message_queue()[consumer_id]
        # no consumer or queue is growing too much -> stop consumer and queue
        if (not cb_info or not cb_info['callback']) or len(pending_messages) > MAX_PENDING_MESSAGES_PER_CONSUMER:
            consumers_callbacks().pop(consumer_id, None)
            message_queue().pop(consumer_id, None)
            if _Debug:
                lg.out(_DebugLevel, 'message.do_read STOPPED consumer "%s", pending_messages=%d' % (consumer_id, len(pending_messages)))
            continue
        # filter messages which consumer is not interested in
        if cb_info['direction']:
            consumer_messages = filter(lambda msg: msg['dir'] == cb_info['direction'], pending_messages)
        else:
            consumer_messages = filter(None, pending_messages)
        if cb_info['message_types']:
            consumer_messages = filter(lambda msg: msg['type'] in cb_info['message_types'], consumer_messages)
        consumer_messages = list(consumer_messages)
        if not consumer_messages:
            message_queue()[consumer_id] = []
            continue
        # callback is a one-time Deferred object, must call it now and release the callback
        if isinstance(cb_info['callback'], Deferred):
            if cb_info['callback'].called:
                if _Debug:
                    lg.out(_DebugLevel, 'message.do_read %d messages waiting consuming by "%s", callback state is "called"' % (len(message_queue()[consumer_id]), consumer_id))
                consumers_callbacks().pop(consumer_id, None)
                continue
            try:
                cb_info['callback'].callback(consumer_messages)
            except:
                lg.exc()
                consumers_callbacks().pop(consumer_id, None)
                continue
            consumers_callbacks().pop(consumer_id, None)
            message_queue()[consumer_id] = []
            total_handled += len(consumer_messages)
            continue
        # callback is a "callable" method which we must not release
        message_queue()[consumer_id] = []
        try:
            handled = cb_info['callback'](consumer_messages)
        except:
            lg.exc()
            consumers_callbacks().pop(consumer_id, None)
            # put back failed messages to the queue so consumer can re-try
            message_queue()[consumer_id] = pending_messages
            continue
        if _Debug:
            lg.args(_DebugLevel, handled=handled, cb_info=cb_info)
        if handled is None:
            lg.err('failed consuming messages by consumer %r' % consumer_id)
            consumers_callbacks().pop(consumer_id, None)
            # put back failed messages to the queue so consumer can re-try
            message_queue()[consumer_id] = pending_messages
            continue
        if handled:
            total_handled += len(consumer_messages)
    if _Debug:
        lg.args(_DebugLevel, total_handled=total_handled)
    return total_handled



import os
import sys
import time
import hashlib
import threading

from decimal import Decimal

import essentials
import mining_heavy3
import regnet
import mempool as mp
import tokensv2 as tokens
from difficulty import difficulty
from essentials import address_is_rsa, checkpoint_set, ledger_balance3
from polysign.signerfactory import SignerFactory
from quantizer import quantize_two, quantize_eight
from fork import Fork

fork = Fork()


def digest_block(node, data, sdef, peer_ip, db_handler):

    # if not mining_heavy3.MMAP:
    #     mining_heavy3.mining_open(node.heavy3_path)

    """node param for imports"""
    class Transaction:
        def __init__(self):
            self.start_time_tx = 0
            self.q_received_timestamp = 0
            self.received_timestamp = '0.00'
            self.received_address = None
            self.received_recipient = None
            self.received_amount = 0
            self.received_signature_enc = None
            self.received_public_key_b64encoded = None
            self.received_operation = None
            self.received_openfield = None

    class MinerTransaction:
        def __init__(self):
            self.q_block_timestamp = 0
            self.nonce = None
            self.miner_address = None

    class Block:
        """array of transactions within a block"""
        def __init__(self):
            self.tx_count = 0
            self.block_height_new = node.last_block + 1
            self.block_hash = 'N/A'
            self.failed_cause = ''
            self.block_count = 0
            self.transaction_list_converted = []

            self.mining_reward = None
            self.mirror_hash = None
            self.start_time_block = quantize_two(time.time())
            self.tokens_operation_present = False

    def fork_reward_check():
        # fork handling
        if node.is_testnet:
            if node.last_block > fork.POW_FORK_TESTNET:
                if not fork.check_postfork_reward_testnet(db_handler):
                    db_handler.rollback_under(fork.POW_FORK_TESTNET - 1)
                    raise ValueError('Rolling back chain due to old fork data')
        else:
            if node.last_block > fork.POW_FORK:
                if not fork.check_postfork_reward(db_handler):
                    # print('Rolling back')
                    db_handler.rollback_under(fork.POW_FORK - 1)
                    raise ValueError('Rolling back chain due to old fork data')
        # fork handling

    def transaction_validate():
        """Validates all transaction elements. Raise a ValueError exception on error."""

        # Begin with costless checks first, so we can early exit. Time of tx
        if tx.start_time_tx < tx.q_received_timestamp:
            raise ValueError(f'Future transaction not allowed, timestamp '
                             f'{quantize_two((tx.q_received_timestamp - tx.start_time_tx) / 60)} minutes in the future', )
        if node.last_block_timestamp - 86400 > tx.q_received_timestamp:
            raise ValueError('Transaction older than 24h not allowed.')
        # Amount
        if float(tx.received_amount) < 0:
            raise ValueError('Negative balance spend attempt')
        # Addresses validity
        if not essentials.address_validate(tx.received_address):
            raise ValueError('Not a valid sender address')
        if not essentials.address_validate(tx.received_recipient):
            raise ValueError('Not a valid recipient address')

        # Now we can process cpu heavier checks, decode and check sig itself
        buffer = str((tx.received_timestamp, tx.received_address, tx.received_recipient, tx.received_amount, tx.received_operation, tx.received_openfield)).encode('utf-8')
        # Will raise if error - also includes reconstruction of address from pubkey to make sure it matches
        SignerFactory.verify_bis_signature(tx.received_signature_enc, tx.received_public_key_b64encoded, buffer, tx.received_address)
        node.logger.app_log.info(f'Valid signature from {tx.received_address} '
                                 f'to {tx.received_recipient} amount {tx.received_amount}', )

    def rewards():
        if int(block_instance.block_height_new) % 10 == 0:  # every 10 blocks
            db_handler.dev_reward(node, block_instance, miner_tx, block_instance.mining_reward, block_instance.mirror_hash)
            db_handler.hn_reward(node, block_instance, miner_tx, block_instance.mirror_hash)

    def check_signature(block):
        # TODO EGG: benchmark this loop vs a single "WHERE IN" SQL
        signature_list = []

        for entry in block:  # sig 4

            entry_signature = entry[4]
            if entry_signature:  # prevent empty signature database retry hack
                signature_list.append(entry_signature)
                # reject block with transactions which are already in the ledger ram
                if node.old_sqlite:
                    db_handler.execute_param(db_handler.h, 'SELECT block_height FROM transactions WHERE signature = ?1;', (entry_signature, ))
                else:
                    db_handler.execute_param(db_handler.h, 'SELECT block_height FROM transactions WHERE substr(signature,1,4) = substr(?1,1,4) and signature = ?1;', (entry_signature, ))

                tx_presence_check = db_handler.h.fetchone()
                if tx_presence_check:
                    # print(node.last_block)
                    raise ValueError(f'That transaction {entry_signature[:10]} is already in our ledger, '
                                     f'block_height {tx_presence_check[0]}', )
                if node.old_sqlite:
                    db_handler.execute_param(db_handler.c, 'SELECT block_height FROM transactions WHERE signature = ?1;', (entry_signature, ))
                else:
                    db_handler.execute_param(db_handler.c, 'SELECT block_height FROM transactions WHERE substr(signature,1,4) = substr(?1,1,4) and signature = ?1;', (entry_signature, ))
                tx_presence_check = db_handler.c.fetchone()
                if tx_presence_check:
                    # print(node.last_block)
                    raise ValueError(f'That transaction {entry_signature[:10]} is already in our RAM ledger, '
                                     f'block_height {tx_presence_check[0]}', )
            else:
                raise ValueError(f'Empty signature from {peer_ip}')

        if block_instance.tx_count != len(set(signature_list)):
            raise ValueError('There are duplicate transactions in this block, rejected')

    def sort_transactions(block):
        # print("sort_transactions")
        # print("block_instance.tx_count", block_instance.tx_count)
        for tx_index, transaction in enumerate(block):
            # print("tx_index", tx_index)
            tx.start_time_tx = quantize_two(time.time())
            tx.q_received_timestamp = quantize_two(transaction[0])
            tx.received_timestamp = '%.2f' % tx.q_received_timestamp
            tx.received_address = str(transaction[1])[:56]
            tx.received_recipient = str(transaction[2])[:56]
            tx.received_amount = '%.8f' % (quantize_eight(transaction[3]))
            tx.received_signature_enc = str(transaction[4])[:684]
            tx.received_public_key_b64encoded = str(transaction[5])[:1068]
            tx.received_operation = str(transaction[6])[:30]
            tx.received_openfield = str(transaction[7])[:100000]

            if tx.received_operation in ['token:issue', 'token:transfer']:
                block_instance.tokens_operation_present = True  # update on change

            # if transaction == block[-1]:
            if tx_index == block_instance.tx_count - 1:  # faster than comparing the whole tx
                if float(tx.received_amount) != 0:
                    raise ValueError('Coinbase (Mining) transaction must have zero amount')
                if not address_is_rsa(tx.received_address):
                    # Compare address rather than sig, as sig could be made up
                    raise ValueError('Coinbase (Mining) transaction only supports legacy RSA Bismuth addresses')

                # recognize the last transaction as the mining reward transaction
                miner_tx.q_block_timestamp = tx.q_received_timestamp
                miner_tx.nonce = tx.received_openfield[:128]
                miner_tx.miner_address = tx.received_address
                # print("miner_tx1", miner_tx)

            block_instance.transaction_list_converted.append(
                (tx.received_timestamp, tx.received_address, tx.received_recipient, tx.received_amount, tx.received_signature_enc, tx.received_public_key_b64encoded, tx.received_operation, tx.received_openfield)
            )
            transaction_validate()

    def process_transactions(block):
        try:
            fees_block = []
            block_instance.mining_reward = 0  # avoid warning

            # Cache for multiple tx from same address
            balances = {}

            # TODO: remove condition after HF
            if block_instance.block_height_new >= 1450000:
                oldest_possible_tx = miner_tx.q_block_timestamp - 60*60*2
            else:
                # Was 24 h before
                oldest_possible_tx = miner_tx.q_block_timestamp - 60*60*24

            for tx_index, transaction in enumerate(block):
                if float(transaction[0]) < oldest_possible_tx:
                    raise ValueError('txid {} from {} is older ({}) than oldest possible date ({})'.format(transaction[4][:56], transaction[1], transaction[0], oldest_possible_tx))
                db_timestamp = '%.2f' % quantize_two(transaction[0])
                db_address = str(transaction[1])[:56]
                db_recipient = str(transaction[2])[:56]
                db_amount = '%.8f' % quantize_eight(transaction[3])
                db_signature = str(transaction[4])[:684]
                db_public_key_b64encoded = str(transaction[5])[:1068]
                db_operation = str(transaction[6])[:30]
                db_openfield = str(transaction[7])[:100000]

                block_debit_address = 0
                block_fees_address = 0

                # this also is redundant on many tx per address block
                for x in block:
                    if x[1] == db_address:  # make calculation relevant to a particular address in the block
                        block_debit_address = quantize_eight(Decimal(block_debit_address) + Decimal(x[3]))

                        if x != block[-1]:
                            block_fees_address = quantize_eight(Decimal(block_fees_address) + Decimal(essentials.fee_calculate(db_openfield, db_operation, node.last_block)))  # exclude the mining tx from fees

                # node.logger.app_log.info("Fee: " + str(fee))

                # decide reward
                if tx_index == block_instance.tx_count - 1:
                    db_amount = 0  # prevent spending from another address, because mining txs allow delegation

                    # if node.is_testnet and node.last_block >= fork.POW_FORK_TESTNET:
                    #     block_instance.mining_reward = 15 - (block_instance.block_height_new - fork.POW_FORK_TESTNET)/1100000 - 9.5
                    # elif node.is_mainnet and node.last_block >= fork.POW_FORK:
                    #     block_instance.mining_reward = 15 - (block_instance.block_height_new - fork.POW_FORK)/1100000 - 9.5
                    # else:
                    #     block_instance.mining_reward = 15 - (quantize_eight(block_instance.block_height_new)/quantize_eight(1000000/2)) - Decimal('2.4')
                    # if block_instance.mining_reward < 0.5:
                    #    block_instance.mining_reward = 0.5
                    block_instance.mining_reward = node.REGULAR_MINER_REWARD
                    if miner_tx.miner_address in node.FOUNDATION_MINERS and block_instance.tx_count == 1:
                        block_instance.mining_reward = node.FOUNDATION_MINER_REWARD

                    reward = '{:.8f}'.format(Decimal(block_instance.mining_reward) + sum(fees_block))

                    # don't request a fee for mined block so new accounts can mine
                    fee = 0
                else:
                    reward = 0
                    fee = essentials.fee_calculate(db_openfield, db_operation, node.last_block)
                    fees_block.append(quantize_eight(fee))
                    balance_pre = ledger_balance3(db_address, balances, db_handler)  # keep this as c (ram hyperblock access)
                    balance = quantize_eight(balance_pre - block_debit_address)

                    if quantize_eight(balance_pre) < quantize_eight(db_amount):
                        raise ValueError(f'{db_address} sending more than owned: {db_amount}/{balance_pre}')

                    if quantize_eight(balance) - quantize_eight(block_fees_address) < 0:
                        # exclude fee check for the mining/header tx
                        raise ValueError(f'{db_address} Cannot afford to pay fees (balance: {balance}, '
                                         f'block fees: {block_fees_address})', )

                # append, but do not insert to ledger before whole block is validated,
                # note that it takes already validated values (decimals, length)
                node.logger.app_log.info(f'Chain: Appending transaction back to block with '
                                         f'{len(block_transactions)} transactions in it', )
                block_transactions.append(
                    (
                        str(block_instance.block_height_new),
                        str(db_timestamp),
                        str(db_address),
                        str(db_recipient),
                        str(db_amount),
                        str(db_signature),
                        str(db_public_key_b64encoded),
                        str(block_instance.block_hash),
                        str(fee),
                        str(reward),
                        str(db_operation),
                        str(db_openfield),
                    )
                )
                try:
                    mp.MEMPOOL.delete_transaction(db_signature)
                    node.logger.app_log.info(f'Chain: Removed processed transaction {db_signature[:56]}'
                                             f' from the mempool while digesting', )
                except:
                    # tx was not or is no more in the local mempool
                    pass
        except Exception as e:
            print(e)
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            raise

    def process_blocks(b_data):
        # TODO: block_data shadows block_data from outer scope. Very error prone.
        # here, functions in functions use both local vars or parent variables, it's a call for nasty bugs.
        # take care of pycharms hints, do not define func in funcs.
        try:
            block_instance.block_count = len(b_data)

            for block in b_data:
                if node.IS_STOPPING:
                    node.logger.app_log.warning('Process_blocks aborted, node is stopping')
                    return
                # Reworked process: we exit as soon as we find an error, no need to process further tests.
                # Then the exception handler takes place.
                # EGG: Reminder: quick test first, **always**. Heavy tests only thereafter.

                block_instance.tx_count = len(block)

                block_instance.block_height_new = node.last_block + 1
                block_instance.start_time_block = quantize_two(time.time())

                # fork_reward_check()

                # sort_transactions also computes several hidden variables, like miner_tx.q_block_timestamp
                # So it has to be run before the check
                # TODO: rework to avoid hidden variables and make the sequence clear.
                # sort_transactions also validates all transactions and sigs, and this is a waste of time if the block timestamp is wrong.
                sort_transactions(block)
                # reject blocks older than latest block
                if miner_tx.q_block_timestamp <= node.last_block_timestamp:
                    # print("miner_tx2", miner_tx)
                    raise ValueError(f'!Block is older {miner_tx.q_block_timestamp} '
                                     f'than the previous one {node.last_block_timestamp} , will be rejected', )

                if len(block) == 1:
                    if miner_tx.miner_address not in node.FOUNDATION_MINERS:
                        transaction = block[0]
                        received_operation = str(transaction[6])[:30]
                        received_amount = float(transaction[3])
                        if received_operation.startswith('identity') and received_amount == 0:
                            pass
                        else:
                            # only allow foundation miners to mine empty blocks
                            raise ValueError('Only foundation miners are allowed to mine empty blocks')

                check_signature(block)

                # calculate current difficulty (is done for each block in block array, not super easy to isolate)
                diff = difficulty(node, db_handler)
                node.difficulty = diff

                # node.logger.app_log.warning(f"Time to generate block {node.last_block + 1}: {'%.2f' % diff[2]}")
                # node.logger.app_log.warning(f'Current difficulty: {diff[3]}')
                # node.logger.app_log.warning(f'Current blocktime: {diff[4]}')
                # node.logger.app_log.warning(f'Current hashrate: {diff[5]}')
                # node.logger.app_log.warning(f'Difficulty adjustment: {diff[6]}')
                # node.logger.app_log.warning(f'Difficulty: {diff[0]} {diff[1]}')

                block_instance.block_hash = hashlib.sha224((str(block_instance.transaction_list_converted) + node.last_block_hash).encode('utf-8')).hexdigest()
                del block_instance.transaction_list_converted[:]

                # node.logger.app_log.info("Last block sha_hash: {}".format(block_hash))
                # node.logger.app_log.info(f'Calculated block sha_hash: {block_instance.block_hash}')
                # node.logger.app_log.info("Nonce: {}".format(nonce))

                # check if we already have the sha_hash
                db_handler.execute_param(db_handler.h, 'SELECT block_height FROM transactions WHERE block_hash = ?', (block_instance.block_hash, ))
                dummy = db_handler.h.fetchone()
                if dummy:
                    raise ValueError('Skipping digestion of block {} from {}, because we already have it on block_height {}'.format(block_instance.block_hash[:10], peer_ip, dummy[0]))

                mining_heavy3.mining_open(node.heavy3_path)
                if node.is_mainnet:
                    diff_save = mining_heavy3.check_block(
                        block_instance.block_height_new,
                        miner_tx.miner_address,
                        miner_tx.nonce,
                        node.last_block_hash,
                        diff[0],
                        tx.received_timestamp,
                        tx.q_received_timestamp,
                        node.last_block_timestamp,
                        peer_ip=peer_ip,
                        app_log=node.logger.app_log,
                    )
                elif node.is_testnet:
                    diff_save = mining_heavy3.check_block(
                        block_instance.block_height_new,
                        miner_tx.miner_address,
                        miner_tx.nonce,
                        node.last_block_hash,
                        diff[0],
                        tx.received_timestamp,
                        tx.q_received_timestamp,
                        node.last_block_timestamp,
                        peer_ip=peer_ip,
                        app_log=node.logger.app_log,
                    )
                else:
                    # it's regnet then, will use a specific fake method here.
                    diff_save = mining_heavy3.check_block(
                        block_instance.block_height_new,
                        miner_tx.miner_address,
                        miner_tx.nonce,
                        node.last_block_hash,
                        regnet.REGNET_DIFF,
                        tx.received_timestamp,
                        tx.q_received_timestamp,
                        node.last_block_timestamp,
                        peer_ip=peer_ip,
                        app_log=node.logger.app_log,
                    )

                process_transactions(block)

                node.last_block = block_instance.block_height_new
                node.last_block_hash = block_instance.block_hash
                # end for block

                # save current diff (before the new block)

                # quantized vars have to be converted, since Decimal is not json serializable...
                node.plugin_manager.execute_action_hook(
                    'block',
                    {
                        'height': block_instance.block_height_new,
                        'diff': diff_save,
                        'hash': block_instance.block_hash,
                        'timestamp': float(miner_tx.q_block_timestamp),
                        'miner': miner_tx.miner_address,
                        'ip': peer_ip,
                    },
                )

                node.plugin_manager.execute_action_hook(
                    'fullblock',
                    {
                        'height': block_instance.block_height_new,
                        'diff': diff_save,
                        'hash': block_instance.block_hash,
                        'timestamp': float(miner_tx.q_block_timestamp),
                        'miner': miner_tx.miner_address,
                        'ip': peer_ip,
                        'transactions': block_transactions,
                    },
                )

                db_handler.to_db(block_instance, diff_save, block_transactions)

                # new sha_hash
                db_handler.execute(
                    db_handler.c,
                    'SELECT * FROM transactions '
                    'WHERE block_height = (SELECT max(block_height) FROM transactions)',
                )
                # Was trying to simplify, but it's the latest mirror sha_hash.
                # not the latest block, nor the mirror of the latest block.
                # c.execute("SELECT * FROM transactions WHERE block_height = ?", (block_instance.block_height_new -1,))
                tx_list_to_hash = db_handler.c.fetchall()
                block_instance.mirror_hash = hashlib.blake2b(str(tx_list_to_hash).encode(), digest_size=20).hexdigest()
                # /new sha_hash

                # Disabling rewards
                # rewards()

                # node.logger.app_log.warning("Block: {}: {} valid and saved from {}"
                # .format(block_instance.block_height_new, block_hash[:10], peer_ip))
                node.logger.app_log.warning(
                    f'Valid block: {block_instance.block_height_new}: '
                    f'{block_instance.block_hash[:10]} with {len(block)} txs, '
                    f'digestion from {peer_ip} completed in '
                    f'{str(time.time() - float(block_instance.start_time_block))[:5]}s.',
                )

                if block_instance.tokens_operation_present:
                    tokens.tokens_update(node, db_handler)

                del block_transactions[:]
                node.peers.unban(peer_ip)

                # This new block may change the int(diff). Trigger the hook whether it changed or not.
                diff = difficulty(node, db_handler)
                node.difficulty = diff
                node.plugin_manager.execute_action_hook('diff', diff[0])
                # We could recalc diff after inserting block, and then only trigger the block hook,
                # but I fear this would delay the new block event.

                # /whole block validation
                # NEW: returns new block sha_hash
        except Exception as e:
            # Left for edge cases debug
            print(e)
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            raise

    #--- digest start
    # TODO: no def in def, unreadable. we are 10 screens down the prototype of that function.
    # digestion begins here
    if node.peers.is_banned(peer_ip):
        # no need to loose any time with banned peers
        raise ValueError('Cannot accept blocks from a banned peer')
        # since we raise, it will also drop the connection, it's fine since he's banned.

    tx = Transaction()
    miner_tx = MinerTransaction()
    block_instance = Block()

    if not node.db_lock.locked():

        node.db_lock.acquire()
        node.logger.app_log.warning(f'Database lock acquired in {threading.current_thread().name}')

        while mp.MEMPOOL.lock.locked():
            time.sleep(0.1)
            node.logger.app_log.info(f'Chain: Waiting for mempool to unlock {peer_ip}')

        node.logger.app_log.warning(f'Chain: Digesting started from {peer_ip}')
        # variables that have been quantized are prefixed by q_ So we can avoid any unnecessary quantize again later.
        # Takes time. Variables that are only used as quantized decimal are quantized once and for all.

        block_size = Decimal(sys.getsizeof(str(data)))/Decimal(1000000)
        node.logger.app_log.warning(f'Chain: Block size: {block_size} MB')

        try:
            block_data = data
            # reject block with duplicate transactions
            block_transactions = []

            process_blocks(block_data)

            checkpoint_set(node)
            return node.last_block_hash

        except Exception as e:
            node.logger.app_log.warning(f'Chain processing failed: {e}')
            node.logger.app_log.info(f'Received data dump: {data}')
            block_instance.failed_cause = str(e)

            node.last_block = db_handler.block_max_ram()['block_height']  #get actual data from database on exception
            node.last_block_hash = db_handler.last_block_hash()  #get actual data from database on exception

            # Temp
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)

            if node.peers.warning(sdef, peer_ip, 'Rejected block', 2):
                raise ValueError(f'{peer_ip} banned')
            raise ValueError('Chain: digestion aborted')

        finally:

            db_handler.db_to_drive(node)

            node.db_lock.release()
            node.logger.app_log.warning(f'Database lock released in {threading.current_thread().name}')

            delta_t = time.time() - float(block_instance.start_time_block)
            # node.logger.app_log.warning("Block: {}: {} digestion completed in {}s."
            # .format(block_instance.block_height_new,  block_hash[:10], delta_t))
            node.plugin_manager.execute_action_hook('digestblock', {'failed': block_instance.failed_cause, 'ip': peer_ip, 'deltat': delta_t, 'blocks': block_instance.block_count, 'txs': block_instance.tx_count})

    else:
        node.logger.app_log.warning(f'Chain: Skipping processing from {peer_ip}, someone delivered data faster')
        node.plugin_manager.execute_action_hook('digestblock', {'failed': 'skipped', 'ip': peer_ip})
    #--- digest end



#!/usr/bin/env python
# customer_assistant.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (customer_assistant.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: customer_assistant
.. role:: red

BitDust customer_assistant() Automat

EVENTS:
    * :red:`ack`
    * :red:`connect`
    * :red:`disconnect`
    * :red:`fail`
    * :red:`init`
    * :red:`propagate`
    * :red:`shutdown`
    * :red:`timer-10sec`
    * :red:`timer-5min`


TODO: periodically send Request service Customer to be sure that i am still your supplier
if not - remove that customer and stop assistant

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from bitdust.automats import automat

from bitdust.logs import lg

from bitdust.lib import packetid
from bitdust.lib import diskspace

from bitdust.main import settings

from bitdust.crypt import my_keys

from bitdust.p2p import p2p_service
from bitdust.p2p import commands

from bitdust.supplier import list_files

from bitdust.storage import accounting

#------------------------------------------------------------------------------

_CustomerAssistants = {}

#------------------------------------------------------------------------------


def assistants():
    global _CustomerAssistants
    return _CustomerAssistants


def create(customer_idurl):
    if customer_idurl in assistants():
        raise Exception('another CustomerAssistant instance for %s already exists' % customer_idurl)
    assistants()[customer_idurl] = CustomerAssistant(customer_idurl)
    return assistants()[customer_idurl]


def by_idurl(customer_idurl):
    return assistants().get(customer_idurl, None)


#------------------------------------------------------------------------------


class CustomerAssistant(automat.Automat):
    """
    This class implements all the functionality of the ``customer_assistant()`` state machine.
    """

    timers = {
        'timer-10sec': (10.0, ['PING?']),
        'timer-5min': (300, ['CONNECTED']),
    }

    def __init__(self, customer_idurl):
        """
        Create customer_assistant() state machine for given customer.
        """
        self.customer_idurl = customer_idurl
        self.donated_bytes = accounting.get_customer_quota(self.customer_idurl) or 0
        name = 'customer_%s_%s' % (
            self.customer_idurl.to_id(),
            diskspace.MakeStringFromBytes(self.donated_bytes).replace(' ', ''),
        )
        super(CustomerAssistant, self).__init__(
            name=name,
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=False,
            log_transitions=_Debug,
        )

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <https://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'OFFLINE'
                self.doInit(*args, **kwargs)
        #---OFFLINE---
        elif self.state == 'OFFLINE':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'connect':
                self.state = 'PING?'
                self.doSendMyIdentity(*args, **kwargs)
            elif event == 'propagate':
                self.state = 'PING?'
        #---PING?---
        elif self.state == 'PING?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'ack':
                self.state = 'CONNECTED'
                self.doSendHisFiles(*args, **kwargs)
            elif event == 'timer-10sec' or event == 'disconnect' or event == 'fail':
                self.state = 'OFFLINE'
        #---CONNECTED---
        elif self.state == 'CONNECTED':
            if event == 'disconnect':
                self.state = 'OFFLINE'
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'timer-5min':
                self.state = 'PING?'
                self.doSendMyIdentity(*args, **kwargs)
            elif event == 'propagate':
                self.state = 'PING?'
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """

    def doSendMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        p2p_service.SendIdentity(
            self.customer_idurl,
            wide=True,
            callbacks={
                commands.Ack(): self._customer_acked,
                commands.Fail(): self._customer_failed,
            },
        )

    def doSendHisFiles(self, *args, **kwargs):
        """
        Action method.
        """
        #         customer_key_id = my_keys.make_key_id(alias='customer', creator_idurl=self.customer_idurl)
        #         customer_key_id = my_keys.latest_key_id(customer_key_id)
        # if my_keys.is_key_registered(customer_key_id):
        # TODO: re-think again about the customer key, do we really need it?
        #         if False:
        #             list_files.send(
        #                 customer_idurl=self.customer_idurl,
        #                 packet_id='%s:%s' % (
        #                     customer_key_id,
        #                     packetid.UniqueID()
        #                 ),
        #                 format_type=settings.ListFilesFormat(),
        #                 key_id=customer_key_id,
        #                 remote_idurl=self.customer_idurl,  # send to the customer
        #             )
        #         else:
        #             # if "customer" key is not delivered to me yet, use his "master" key
        customer_master_key_id = my_keys.latest_key_id(my_keys.make_key_id(alias='master', creator_idurl=self.customer_idurl))
        if my_keys.is_key_registered(customer_master_key_id):
            list_files.send(
                customer_idurl=self.customer_idurl,
                packet_id='%s:%s' % (customer_master_key_id, packetid.UniqueID()),
                format_type=settings.ListFilesFormat(),
                key_id=customer_master_key_id,
                remote_idurl=self.customer_idurl,
            )
        else:
            lg.err('key %s is not registered, not able to send customer files' % customer_master_key_id)

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        assistants().pop(self.customer_idurl)
        self.destroy()

    def _customer_acked(self, response, info):
        if _Debug:
            lg.out(_DebugLevel, 'customer_assistant._customer_acked %r %r' % (response, info))
        self.automat(response.Command.lower(), response)

    def _customer_failed(self, response, info):
        if _Debug:
            lg.out(_DebugLevel, 'customer_assistant._customer_failed %r %r' % (response, info))
        event_id = 'fail'
        if response:
            event_id = response.Command.lower()
        self.automat(event_id, response)



"""
A Tornado based wallet server for Bismuth.
EggdraSyl - June 2018


pip3 install -r requirements.txt
"""

import asyncio
import datetime
import json
import logging
import os
import sys
import time
from logging.handlers import RotatingFileHandler

import aioprocessing
import psutil
import tornado.gen
import tornado.log
import tornado.util
from tornado.ioloop import IOLoop
from tornado.iostream import StreamClosedError
from tornado.options import define, options
from tornado.tcpserver import TCPServer

# Bismuth specific modules
import modules.config as config
# from modules.helpers import *
from modules.sqlitebase import SqliteBase
from modules.ledgerbase import LedgerBase
from modules.node_interface import NodeInterface

__version__ = '0.1.22'

# Server
# -----------------------------------------------------------------------------


class WalletServer(TCPServer):
    """Tornado asynchronous TCP server."""
    clients = set()
    status_dict = {'version': __version__}
    node_interface = None

    async def handle_stream(self, stream, address):
        global access_log
        global app_log
        global stop_event
        global MAX_CLIENTS
        ip, fileno = address
        if len(self.clients) >= MAX_CLIENTS:
            access_log.info('Reached {} max clients, denying connection for {}.'.format(MAX_CLIENTS, ip))
            return
        WalletServer.clients.add(address)
        access_log.info('Incoming connection from {}:{} - {} Total Clients'.format(ip, fileno, len(self.clients)))
        while not stop_event.is_set():
            try:
                # print("waiting for command")
                await self.command(stream, ip)
            except StreamClosedError:
                WalletServer.clients.remove(address)
                access_log.info('Client {}:{} left  - {} Total Clients'.format(ip, fileno, len(self.clients)))
                break
            except ValueError:
                WalletServer.clients.remove(address)
                if options.verbose:
                    access_log.info('Client {}:{} Rejected  - {} Total Clients'.format(ip, fileno, len(self.clients)))
                stream.close()
                break
            except tornado.util.TimeoutError:
                WalletServer.clients.remove(address)
                access_log.info('Client {}:{} Timeout  - {} Total Clients'.format(ip, fileno, len(self.clients)))
                try:
                    stream.close()
                except Exception:
                    pass
                break
            except Exception as e:
                what = str(e)
                if 'OK' not in what:
                    app_log.error('handle_stream {} for ip {}:{}'.format(what, ip, fileno))
                    await asyncio.sleep(1)

    @staticmethod
    async def _receive(stream, ip):
        """
        Get a command, async version
        :param stream:
        :param ip:
        :return:
        """
        try:
            header = await tornado.gen.with_timeout(datetime.timedelta(seconds=35), stream.read_bytes(10), quiet_exceptions=tornado.iostream.StreamClosedError)
            data_len = int(header)
            data = await tornado.gen.with_timeout(datetime.timedelta(seconds=10), stream.read_bytes(data_len), quiet_exceptions=tornado.iostream.StreamClosedError)
            data = json.loads(data.decode('utf-8'))
            return data
        except Exception as e:
            app_log.error('_receive {} for ip {}'.format(str(e), ip))
            raise

    @staticmethod
    async def _send(data, stream, ip):
        """
        sends an object to the stream, async.
        :param data:
        :param stream:
        :param ip:
        :return:
        """
        global app_log
        try:
            data = str(json.dumps(data))
            header = str(len(data)).encode('utf-8').zfill(10)
            full = header + data.encode('utf-8')
            await stream.write(full)
        except Exception as e:
            app_log.error('_send {} for ip {}'.format(str(e), ip))
            raise

    async def command(self, stream, ip):
        global access_log
        data = await self._receive(stream, ip)
        if options.verbose:
            access_log.info('Command ' + data)

        # roles
        rights = await getrights(ip)

        if data == 'wstatusget':
            # Get wallet server status only
            await self._send(self.status_dict, stream, ip)
            return

        # Other commands are automagically deduced from the node interface.

        # how many params to get for that command?
        params_count = self.node_interface.param_count_of(data, rights)
        print('{} params'.format(params_count))
        if params_count == -1:
            raise ValueError('Unknown Command ' + data)

        params = [data]
        for i in range(params_count):
            param = await self._receive(stream, ip)
            params.append(param)
        res = await self.node_interface.call_user(params)
        await self._send(res, stream, ip)
        return
        """
        if data == "mpclear":
            # TODO: only for admin
            return

        # admin only
        if "admin" in rights:
            # TODO
            if data == 'status':
                return
            if data == 'setconfig':
                return
        """

    async def background(self):
        """
        This runs in a background coroutine and print out status
        :return:
        """
        global stop_event
        global app_log
        global process
        while not stop_event.is_set():
            try:
                app_log.info('STATUS: {} Connected clients.'.format(len(self.clients)))
                self.status_dict['clients'] = len(self.clients)
                self.status_dict['max_clients'] = MAX_CLIENTS
                if process:
                    of = len(process.open_files())
                    fd = process.num_fds()
                    co = len(process.connections(kind='tcp4'))
                    self.status_dict['of'], self.status_dict['fd'], self.status_dict['co'] = of, fd, co
                    app_log.info('STATUS: {} Open files, {} connections, {} FD used.'.format(of, co, fd))

                await asyncio.sleep(30)
            except Exception as e:
                app_log.error('Error background {}'.format(str(e)))


async def getrights(ip):
    global app_log
    global CONFIG
    try:
        result = ['none']
        if ip in CONFIG.allowed:
            result.append('admin')
        return result
    except Exception as e:
        app_log.warning('Error getrights {}'.format(str(e)))


def start_server(port):
    global app_log
    global stop_event
    global PORT
    global CONFIG
    mempool = SqliteBase(options.verbose, db_path=CONFIG.mempool_path.replace('mempool.db', ''), db_name='mempool.db', app_log=app_log)
    db_name = 'ledger.db'
    if CONFIG.testnet:
        db_name = 'test.db'
    ledger = LedgerBase(options.verbose, db_path=CONFIG.db_path + '/', db_name=db_name, app_log=app_log)

    node_interface = NodeInterface(mempool, ledger, CONFIG, app_log=app_log)
    server = WalletServer()
    server.node_interface = node_interface
    # attach mempool db
    # server.listen(port)
    io_loop = IOLoop.instance()

    if CONFIG.direct_ledger:
        try:
            # Force a db connection attempt and updates db version of ledger
            _ = io_loop.run_sync(ledger.check_db_version, 30)
        except Exception as e:
            app_log.error("Can't connect to ledger: {}".format(e))
            return
    else:
        app_log.info("Config: don't use direct ledger access")
    try:
        # Force a db connection attempt
        _ = io_loop.run_sync(mempool.schema, 30)
    except Exception as e:
        app_log.info("Can't connect to mempool: {}".format(e))
        return

    server.bind(port)
    server.start(1)  # Force one process only
    if options.verbose:
        app_log.info('Starting server on tcp://localhost:{}'.format(port))
    io_loop.spawn_callback(server.background)
    try:
        io_loop.start()
    except KeyboardInterrupt:
        stop_event.set()
        io_loop.stop()
        app_log.info('exited cleanly')


if __name__ == '__main__':
    CONFIG = config.Get()
    CONFIG.read('wallet_server.txt', 'wallet_server_custom.txt')
    #
    is_testnet = CONFIG.testnet
    PORT = CONFIG.port
    MAX_CLIENTS = CONFIG.max_clients

    define('port', default=PORT, help='port to listen on')
    define('verbose', default=False, help='verbose')
    options.parse_command_line()

    # TODO
    """
    if CONFIG.mempool_ram_conf:
        print("Incompatible setting detected.")
        print("Please edit config.txt, set mempool_ram_conf=False and restart node")
        sys.exit()
    """

    # TODO: print settings

    if not os.path.isfile(CONFIG.mempool_path):
        print('mempool.db not found at {}'.format(CONFIG.mempool_path))
        print("Please edit node's config.txt, check mempool_ram_conf=False and restart node.")
        sys.exit()

    start_time = time.time()

    stop_event = aioprocessing.AioEvent()  # Event()

    lock = aioprocessing.AioLock()

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.INFO)
    # formatter = logging.Formatter('%(asctime)s %(funcName)s(%(lineno)d) %(message)s')
    # ch.setFormatter(formatter)
    app_log = logging.getLogger('tornado.application')
    tornado.log.enable_pretty_logging()
    # app_log.addHandler(ch)
    logfile = os.path.abspath('wallet_app.log')
    # Rotate log after reaching 512K, keep 5 old copies.
    rotateHandler = RotatingFileHandler(logfile, 'a', 512*1024, 10)
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    rotateHandler.setFormatter(formatter)
    app_log.addHandler(rotateHandler)

    access_log = logging.getLogger('tornado.access')
    tornado.log.enable_pretty_logging()
    logfile2 = os.path.abspath('wallet_access.log')
    rotateHandler2 = RotatingFileHandler(logfile2, 'a', 512*1024, 10)
    formatter2 = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    rotateHandler2.setFormatter(formatter2)
    access_log.addHandler(rotateHandler2)

    app_log.warning('Testnet: {}'.format(is_testnet))
    # fail safe
    if is_testnet and int(CONFIG.node_port) != 2829:
        app_log.warning('Testnet is active, but node_port set to {} instead of 2829. '
                        'Make sure!'.format(CONFIG.node_port), )
        time.sleep(2)

    if os.name == 'posix':
        process = psutil.Process()
        try:
            # Fails on alpine linux
            limit = process.rlimit(psutil.RLIMIT_NOFILE)
        except Exception:
            limit = (1024, -1)
        app_log.info('OS File limits {}, {}'.format(limit[0], limit[1]))
        if limit[0] < 1024:
            app_log.error('Too small ulimit, please tune your system.')
            sys.exit()
    else:
        process = None

    app_log.info('Wallet Server {} Starting on port {}.'.format(__version__, options.port))

    start_server(options.port)



#!/usr/bin/python
# local_tester.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (local_tester.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: local_tester.

Checks that customer packets on the local disk still have good signatures and are valid.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 8

#------------------------------------------------------------------------------

import os
import sys

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in local_tester.py')

from twisted.internet import threads

#------------------------------------------------------------------------------

try:
    from bitdust.logs import lg
except:
    dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))

from bitdust.system import bpio

from bitdust.main import settings

#-------------------------------------------------------------------------------

_TesterQueue = []
_CurrentProcess = None
_Loop = None
_LoopValidate = None
_LoopUpdateCustomers = None
_LoopSpaceTime = None

#------------------------------------------------------------------------------

TesterUpdateCustomers = 'update_customers'
TesterValidate = 'validate'
TesterSpaceTime = 'space_time'

#------------------------------------------------------------------------------


def init():
    global _Loop
    if _Debug:
        lg.out(_DebugLevel, 'local_tester.init')
    _Loop = reactor.callLater(5, loop)  # @UndefinedVariable


def shutdown():
    global _Loop
    global _CurrentProcess
    if _Debug:
        lg.out(_DebugLevel, 'local_tester.shutdown')

    stop()

    if _Loop:
        if _Loop.active():
            _Loop.cancel()
            _Loop = None

    if alive():
        # TODO: use some simple method to notify bptester thread to stop
        # for example write to local file some simple "marker"
        # bptester suppose to read that file every 1-2 seconds and check if "marker" is here
        if _Debug:
            lg.out(_DebugLevel, 'local_tester.shutdown is killing bptester')


#------------------------------------------------------------------------------


def start():
    global _LoopValidate
    global _LoopUpdateCustomers
    global _LoopSpaceTime
    if _Debug:
        lg.out(_DebugLevel, 'local_tester.start')
    _LoopValidate = reactor.callLater(0, loop_validate)  # @UndefinedVariable
    _LoopUpdateCustomers = reactor.callLater(0, loop_update_customers)  # @UndefinedVariable
    _LoopSpaceTime = reactor.callLater(0, loop_space_time)  # @UndefinedVariable


def stop():
    global _LoopValidate
    global _LoopUpdateCustomers
    global _LoopSpaceTime
    if _Debug:
        lg.out(_DebugLevel, 'local_tester.stop')
    if _LoopValidate:
        if _LoopValidate.active():
            _LoopValidate.cancel()
            _LoopValidate = None
    if _LoopUpdateCustomers:
        if _LoopUpdateCustomers.active():
            _LoopUpdateCustomers.cancel()
            _LoopUpdateCustomers = None
    if _LoopSpaceTime:
        if _LoopSpaceTime.active():
            _LoopSpaceTime.cancel()
            _LoopSpaceTime = None


#------------------------------------------------------------------------------


def _pushTester(cmd):
    global _TesterQueue
    if cmd in _TesterQueue:
        return
    _TesterQueue.append(cmd)


def _popTester():
    global _TesterQueue
    if len(_TesterQueue) == 0:
        return None
    cmd = _TesterQueue[0]
    del _TesterQueue[0]
    return cmd


#-------------------------------------------------------------------------------


def on_thread_finished(ret, cmd):
    global _CurrentProcess
    _CurrentProcess = None
    if _Debug:
        lg.out(_DebugLevel, 'local_tester.on_thread_finished %r with %r' % (cmd, ret))


def run_in_thread(cmd):
    global _CurrentProcess
    from bitdust.main import bptester
    if _CurrentProcess:
        raise Exception('another thread already started')
    _CurrentProcess = cmd
    command = {
        TesterUpdateCustomers: bptester.UpdateCustomers,
        TesterValidate: bptester.Validate,
        TesterSpaceTime: bptester.SpaceTime,
    }[cmd]
    d = threads.deferToThread(command)  # @UndefinedVariable
    d.addBoth(on_thread_finished, cmd)
    if _Debug:
        lg.out(_DebugLevel, 'local_tester.run_in_thread started %r' % cmd)


#------------------------------------------------------------------------------


def alive():
    global _CurrentProcess
    if _CurrentProcess is None:
        return False
    return True


def loop():
    global _Loop
    if not alive():
        cmd = _popTester()
        if cmd:
            run_in_thread(cmd)
    _Loop = reactor.callLater(settings.DefaultLocaltesterLoop(), loop)  # @UndefinedVariable


def loop_validate():
    global _LoopValidate
    TestValid()
    _LoopValidate = reactor.callLater(settings.DefaultLocaltesterValidateTimeout(), loop_validate)  # @UndefinedVariable


def loop_update_customers():
    global _LoopUpdateCustomers
    TestUpdateCustomers()
    _LoopUpdateCustomers = reactor.callLater(settings.DefaultLocaltesterUpdateCustomersTimeout(), loop_update_customers)  # @UndefinedVariable


def loop_space_time():
    global _LoopSpaceTime
    TestSpaceTime()
    _LoopSpaceTime = reactor.callLater(settings.DefaultLocaltesterSpaceTimeTimeout(), loop_space_time)  # @UndefinedVariable


#-------------------------------------------------------------------------------


def TestUpdateCustomers():
    _pushTester(TesterUpdateCustomers)


def TestValid():
    _pushTester(TesterValidate)


def TestSpaceTime():
    _pushTester(TesterSpaceTime)


#-------------------------------------------------------------------------------

if __name__ == '__main__':
    lg.set_debug_level(18)
    bpio.init()
    settings.init()
    init()
    reactor.run()  # @UndefinedVariable
    settings.shutdown()



#!/usr/bin/env python
# accountants_finder.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (accountants_finder.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: accountants_finder.

.. role:: red

BitDust accountants_finder() Automat

EVENTS:
    * :red:`ack-received`
    * :red:`found-one-user`
    * :red:`init`
    * :red:`service-accepted`
    * :red:`service-denied`
    * :red:`shutdown`
    * :red:`start`
    * :red:`timer-10sec`
    * :red:`timer-3sec`
    * :red:`users-not-found`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import strng

from bitdust.p2p import commands
from bitdust.p2p import p2p_service
from bitdust.p2p import lookup

from bitdust.contacts import identitycache

from bitdust.userid import my_id

from bitdust.transport import callback

#------------------------------------------------------------------------------

_AccountantsFinder = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _AccountantsFinder
    if event is None and not args:
        return _AccountantsFinder
    if _AccountantsFinder is None:
        # set automat name and starting state here
        _AccountantsFinder = AccountantsFinder('accountants_finder', 'AT_STARTUP', _DebugLevel, _Debug)
    if event is not None:
        _AccountantsFinder.automat(event, *args, **kwargs)
    return _AccountantsFinder


#------------------------------------------------------------------------------


class AccountantsFinder(automat.Automat):
    """
    This class implements all the functionality of the ``accountants_finder()``
    state machine.
    """

    timers = {
        'timer-3sec': (3.0, ['ACK?']),
        'timer-10sec': (10.0, ['ACK?', 'SERVICE?']),
    }

    def init(self):
        """
        Method to initialize additional variables and flags at creation phase
        of accountants_finder() machine.
        """
        self.target_idurl = None
        self.requested_packet_id = None
        self.request_service_params = None

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python
        <https://bitdust.io/visio2python/>`_ tool.
        """
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'READY'
                self.doInit(*args, **kwargs)
        elif self.state == 'RANDOM_USER':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'found-one-user':
                self.state = 'ACK?'
                self.doRememberUser(*args, **kwargs)
                self.Attempts += 1
                self.doSendMyIdentity(*args, **kwargs)
            elif event == 'users-not-found':
                self.state = 'READY'
                self.doNotifyLookupFailed(*args, **kwargs)
        elif self.state == 'ACK?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'ack-received':
                self.state = 'SERVICE?'
                self.doSendRequestService(*args, **kwargs)
            elif event == 'timer-10sec' and self.Attempts < 5:
                self.state = 'RANDOM_USER'
                self.doLookupRandomUser(*args, **kwargs)
            elif event == 'timer-3sec':
                self.doSendMyIdentity(*args, **kwargs)
        elif self.state == 'SERVICE?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'service-accepted':
                self.state = 'READY'
                self.doNotifyLookupSuccess(*args, **kwargs)
            elif self.Attempts == 5 and (event == 'timer-10sec' or event == 'service-denied'):
                self.state = 'READY'
                self.doNotifyLookupFailed(*args, **kwargs)
            elif (event == 'timer-10sec' or event == 'service-denied') and self.Attempts < 5:
                self.state = 'RANDOM_USER'
                self.doLookupRandomUser(*args, **kwargs)
        elif self.state == 'READY':
            if event == 'start':
                self.state = 'RANDOM_USER'
                self.doSetNotifyCallback(*args, **kwargs)
                self.Attempts = 0
                self.doLookupRandomUser(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        elif self.state == 'CLOSED':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        callback.append_inbox_callback(self._inbox_packet_received)

    def doSetNotifyCallback(self, *args, **kwargs):
        """
        Action method.
        """
        self.result_callback, self.request_service_params = args[0]

    def doLookupRandomUser(self, *args, **kwargs):
        """
        Action method.
        """
        t = lookup.start()
        t.result_defer.addCallback(self._nodes_lookup_finished)
        t.result_defer.addErrback(lambda err: self.automat('users-not-found'))

    def doRememberUser(self, *args, **kwargs):
        """
        Action method.
        """
        self.target_idurl = args[0]

    def doSendMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        p2p_service.SendIdentity(self.target_idurl, wide=True)

    def doSendRequestService(self, *args, **kwargs):
        """
        Action method.
        """
        out_packet = p2p_service.SendRequestService(
            self.target_idurl,
            'service_accountant',
            json_payload=self.request_service_params,
            callbacks={
                commands.Ack(): self._node_acked,
                commands.Fail(): self._node_failed,
            },
        )
        self.requested_packet_id = out_packet.PacketID

    def doNotifyLookupSuccess(self, *args, **kwargs):
        """
        Action method.
        """
        if self.result_callback:
            self.result_callback('accountant-connected', *args, **kwargs)
        self.result_callback = None
        self.request_service_params = None

    def doNotifyLookupFailed(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'accountants_finder.doNotifyLookupFailed, Attempts=%d' % self.Attempts)
        if self.result_callback:
            self.result_callback('lookup-failed', *args, **kwargs)
        self.result_callback = None
        self.request_service_params = None

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        callback.remove_inbox_callback(self._inbox_packet_received)
        self.destroy()
        global _AccountantsFinder
        del _AccountantsFinder
        _AccountantsFinder = None

    #------------------------------------------------------------------------------

    def _inbox_packet_received(self, newpacket, info, status, error_message):
        if newpacket.Command == commands.Ack() and \
                newpacket.OwnerID == self.target_idurl and \
                newpacket.PacketID.startswith('identity:') and \
                self.state == 'ACK?':
            self.automat('ack-received', self.target_idurl)
            return True
        return False

    def _node_acked(self, response, info):
        if _Debug:
            lg.out(_DebugLevel, 'accountants_finder._node_acked %r %r' % (response, info))
        if not strng.to_text(response.Payload).startswith('accepted'):
            if _Debug:
                lg.out(_DebugLevel, 'accountants_finder._node_acked with service denied %r %r' % (response, info))
            self.automat('service-denied')
            return
        if _Debug:
            lg.out(_DebugLevel, 'accountants_finder._node_acked !!!! accountant %s connected' % response.CreatorID)
        self.automat('service-accepted', response.CreatorID)

    def _node_failed(self, response, info):
        if _Debug:
            lg.out(_DebugLevel, 'accountants_finder._node_failed %r %r' % (response, info))
        self.automat('service-denied')

    def _nodes_lookup_finished(self, idurls):
        # TODO: this is still under construction - so I am using this node for tests
        # idurls = ['http://veselin-p2p.ru/bitdust_vps1000_k.xml', ]
        if _Debug:
            lg.out(_DebugLevel, 'accountants_finder._nodes_lookup_finished : %r' % idurls)
        for idurl in idurls:
            ident = identitycache.FromCache(idurl)
            remoteprotos = set(ident.getProtoOrder())
            myprotos = set(my_id.getLocalIdentity().getProtoOrder())
            if len(myprotos.intersection(remoteprotos)) > 0:
                self.automat('found-one-user', idurl)
                return
        self.automat('users-not-found')



#!/usr/bin/python
# broadcast_service.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (broadcast_service.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: broadcast_service.

@author: Veselin
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

# This is used to be able to execute this module directly from command line.
if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

import random
import string
import json

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import utime

from bitdust.crypt import signed
from bitdust.crypt import key

from bitdust.p2p import commands

from bitdust.userid import my_id

#------------------------------------------------------------------------------


def prepare_broadcast_message(owner, payload):
    tm = utime.utcnow_to_sec1970()
    rnd = ''.join(random.choice(string.ascii_uppercase) for _ in range(4))
    msgid = '%s:%s:%s' % (tm, rnd, owner)
    msg = [
        ('owner', owner),
        ('started', tm),
        ('id', msgid),
        ('payload', payload),
    ]
    owner_sign = key.Sign(key.Hash(str(msg)))
    msg = {k: v for k, v in msg}
    msg['owner_sign'] = owner_sign
    return msg


def verfify_broadcast_message(jmsg):
    s = set(jmsg.keys())
    s = s.intersection([
        'owner',
        'started',
        'id',
        'payload',
    ])
    if len(s) != 4:
        return False
    return True


def read_message_from_packet(newpacket):
    try:
        msg = json.loads(newpacket.Payload)
    except:
        lg.exc()
        return None
    # TODO: verify owner signature and creator ID
    return msg


def packet_for_broadcaster(broadcaster_idurl, json_data):
    if 'broadcaster' not in json_data:
        json_data['broadcaster'] = broadcaster_idurl
    return signed.Packet(
        commands.Broadcast(),
        json_data['owner'],
        my_id.getIDURL(),
        json_data['id'],
        json.dumps(json_data),
        broadcaster_idurl,
    )


def packet_for_listener(listener_idurl, json_data):
    # if 'broadcaster' not in json_data:
    json_data['broadcaster'] = my_id.getIDURL()
    return signed.Packet(
        commands.Broadcast(),
        json_data['owner'],
        my_id.getIDURL(),
        json_data['id'],
        json.dumps(json_data),
        listener_idurl,
    )


#------------------------------------------------------------------------------


def send_broadcast_message(payload):
    from broadcast import broadcaster_node
    from broadcast import broadcast_listener
    msg = prepare_broadcast_message(my_id.getIDURL(), payload)
    if broadcaster_node.A():
        broadcaster_node.A('new-outbound-message', (msg, None))
    elif broadcast_listener.A():
        if broadcast_listener.A().state == 'OFFLINE':
            broadcast_listener.A('connect')
        broadcast_listener.A('outbound-message', msg)
    else:
        lg.warn('nor broadcaster_node(), nor broadcast_listener() exists')
        return None
    return msg


#------------------------------------------------------------------------------


def on_incoming_broadcast_message(json_msg):
    lg.out(2, 'service_broadcasting._on_incoming_broadcast_message : %r' % json_msg)


#------------------------------------------------------------------------------


def _test():
    print(prepare_broadcast_message(my_id.getIDURL(), {'test': 'okidoki'}))


#------------------------------------------------------------------------------

if __name__ == '__main__':
    _test()



#!/usr/bin/python
# driver.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (driver.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: driver
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 14

#------------------------------------------------------------------------------

import os
import sys
import importlib

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.defer import Deferred, DeferredList, succeed  # @UnresolvedImport

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.main import config
from bitdust.main import settings
from bitdust.main import events
from bitdust.main import listeners

#------------------------------------------------------------------------------

_Services = {}
_BootUpOrder = []
_RootDistanceDict = {}
_EnabledServices = set()
_DisabledServices = set()
_StartingDeferred = None
_StopingDeferred = None

#------------------------------------------------------------------------------


def services():
    global _Services
    return _Services


def enabled_services():
    global _EnabledServices
    return _EnabledServices


def disabled_services():
    global _DisabledServices
    return _DisabledServices


def boot_up_order():
    global _BootUpOrder
    return _BootUpOrder


def root_distance(service_name):
    global _RootDistanceDict
    return _RootDistanceDict.get(service_name, 0)


def is_on(name):
    svc = services().get(name, None)
    if svc is None:
        return False
    return svc.state == 'ON'


def is_off(name):
    svc = services().get(name, None)
    if svc is None:
        return False
    return svc.state == 'OFF' or svc.state == 'NOT_INSTALLED' or svc.state == 'DEPENDS_OFF'


def is_started(name):
    svc = services().get(name, None)
    if svc is None:
        return False
    return svc.state == 'ON' or svc.state == 'STOPPING' or svc.state == 'STARTING' or svc.state == 'INFLUENCE'


def is_enabled(name):
    svc = services().get(name, None)
    if svc is None:
        return False
    return svc.enabled()


def is_suspended(name):
    svc = services().get(name, None)
    if svc is None:
        return False
    return svc.suspended


def is_exist(name):
    return name in services()


def is_healthy(service_name):
    result = Deferred()
    svc = services().get(service_name, None)
    if svc is None:
        result.errback(Exception('service %s not found' % service_name))
        return result
    if not svc.installed():
        result.errback(Exception('service %s is not installed' % service_name))
        return result
    if not svc.enabled():
        result.errback(Exception('service %s is disabled' % service_name))
        return result
    if not is_started(service_name):
        result.errback(Exception('service %s is not started' % service_name))
        return result
    service_health = svc.health_check()
    if isinstance(service_health, Deferred):
        return service_health
    if service_health is True:
        result.callback(True)
    else:
        result.callback(False)
    return result


def dependent(name):
    svc = services().get(name, None)
    if svc is None:
        return []
    return svc.dependent_on()


def affecting(name):
    svc = services().get(name, None)
    if svc is None:
        return []
    results = set()
    order = list(enabled_services())
    for position in range(len(order)):
        child_name = order[position]
        if child_name == name:
            continue
        child = services()[child_name]
        for depend_name in child.dependent_on():
            if depend_name == name:
                results.add(child_name)
    return list(results)


def request(service_name, service_request_payload, request, info):
    svc = services().get(service_name, None)
    if svc is None:
        raise Exception('service %s not found' % service_name)
    try:
        result = svc.request(service_request_payload, request, info)
    except RequireSubclass:
        from bitdust.p2p import p2p_service
        lg.warn('service %s can not be requested remotely' % service_name)
        return p2p_service.SendFail(request, 'refused')
    except:
        lg.exc()
        return None
    return result


def cancel(service_name, service_cancel_payload, request, info):
    svc = services().get(service_name, None)
    if svc is None:
        raise Exception('service %s not found' % service_name)
    try:
        result = svc.cancel(service_cancel_payload, request, info)
    except RequireSubclass:
        from bitdust.p2p import p2p_service
        lg.warn('service %s can not be cancelled remotely' % service_name)
        return p2p_service.SendFail(request, 'refused')
    except:
        lg.exc()
        return None
    return result


#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'driver.init')
    available_services_dir = os.path.join(bpio.getExecutableDir(), 'services')
    loaded = set()
    for filename in os.listdir(available_services_dir):
        if not filename.endswith('.py') and not filename.endswith('.pyo') and not filename.endswith('.pyc'):
            continue
        if not filename.startswith('service_'):
            continue
        name = str(filename[:filename.rfind('.')])
        if name in loaded:
            continue
        if name in disabled_services():
            if _Debug:
                lg.out(_DebugLevel, '%s is hard disabled' % name)
            continue
        try:
            py_mod = importlib.import_module('bitdust.services.' + name)
        except:
            if _Debug:
                lg.out(_DebugLevel, '%s exception during module import' % name)
            lg.exc()
            continue
        try:
            services()[name] = py_mod.create_service()
        except:
            if _Debug:
                lg.out(_DebugLevel, '%s exception while creating service instance' % name)
            lg.exc()
            continue
        loaded.add(name)
        if not services()[name].enabled():
            if _Debug:
                lg.out(_DebugLevel, '%s is switched off' % name)
            continue
        enabled_services().add(name)
        if _Debug:
            lg.out(_DebugLevel, '%s initialized' % name)
    build_order()
    config.conf().addConfigNotifier('services/', on_service_enabled_disabled)


def shutdown():
    #TODO: rework to make sure all services are properly shutdown and return Deferred object finally
    if _Debug:
        lg.out(_DebugLevel, 'driver.shutdown')
    config.conf().removeConfigNotifier('services/')
    while len(services()):
        name, svc = services().popitem()
        if _Debug:
            lg.out(_DebugLevel, '[%s] CLOSING' % name)
        svc.automat('shutdown')
        del svc
        svc = None
        enabled_services().discard(name)


def build_order():
    global _BootUpOrder
    global _RootDistanceDict
    order = list(enabled_services())
    progress = True
    fail = False
    counter = 0
    while progress and not fail:
        progress = False
        counter += 1
        if counter > len(enabled_services())*len(enabled_services()):
            lg.warn('dependency recursion')
            fail = True
            break
        for position in range(len(order)):
            name = order[position]
            svc = services()[name]
            depend_position_max = -1
            for depend_name in svc.dependent_on():
                if depend_name not in order:
                    fail = True
                    lg.warn('dependency not satisfied: #%d:%s depend on %s' % (position, name, depend_name))
                    break
                depend_position = order.index(depend_name)
                if depend_position > depend_position_max:
                    depend_position_max = depend_position
            if fail:
                break
            if position < depend_position_max:
                # print name, order[depend_position_max]
                order.insert(depend_position_max + 1, name)
                del order[position]
                progress = True
                break
    _BootUpOrder = order
    for service_name in services().keys():
        _RootDistanceDict[service_name] = find_root_distance(service_name)
    return order


def find_root_distance(service_name):
    svc = services().get(service_name)
    if not svc:
        return 0
    if not svc.dependent_on():
        return 0
    max_root_distance = 0
    for depend_name in svc.dependent_on():
        depend_service = services().get(depend_name, None)
        if not depend_service:
            raise ServiceNotFound(depend_name)
        depend_service_root_distance = find_root_distance(depend_name)
        if depend_service_root_distance > max_root_distance:
            max_root_distance = depend_service_root_distance
    return max_root_distance + 1


def start(services_list=[]):
    global _StartingDeferred
    global _StopingDeferred
    if _Debug:
        lg.args(_DebugLevel, services_list=services_list, starting=bool(_StartingDeferred), stoping=bool(_StopingDeferred))
    if _StartingDeferred:
        lg.warn('driver.start already called')
        return _StartingDeferred
    if _StopingDeferred and not _StopingDeferred.called:
        d = Deferred()
        d.errback(Exception('currently another service is stopping'))
        return d
    if not services_list:
        services_list.extend(boot_up_order())
    if _Debug:
        lg.out(_DebugLevel, 'driver.start with %d services' % len(services_list))
    dl = []
    for name in services_list:
        svc = services().get(name, None)
        if not svc:
            raise ServiceNotFound(name)
        if not svc.enabled():
            continue
        if svc.state == 'ON':
            continue
        d = Deferred()
        dl.append(d)
        svc.automat('start', d)
    if len(dl) == 0:
        return succeed(1)
    _StartingDeferred = DeferredList(dl)
    _StartingDeferred.addCallback(on_started_all_services)
    _StartingDeferred.addErrback(on_services_failed_to_start, services_list)
    return _StartingDeferred


def stop(services_list=[]):
    global _StopingDeferred
    global _StartingDeferred
    if _Debug:
        lg.args(_DebugLevel, services_list=services_list, starting=bool(_StartingDeferred), stoping=bool(_StopingDeferred))
    if _StopingDeferred:
        lg.warn('driver.stop already called')
        return _StopingDeferred
    if _StartingDeferred and not _StartingDeferred.called:
        d = Deferred()
        d.errback(Exception('currently another service is starting'))
        return d
    if not services_list:
        services_list.extend(reversed(boot_up_order()))
    if _Debug:
        lg.out(_DebugLevel, 'driver.stop with %d services' % len(services_list))
    dl = []
    for name in services_list:
        svc = services().get(name, None)
        if not svc:
            raise ServiceNotFound(name)
        d = Deferred()
        dl.append(d)
        svc.automat('stop', d)
    _StopingDeferred = DeferredList(dl)
    _StopingDeferred.addCallback(on_stopped_all_services)
    _StopingDeferred.addErrback(on_services_failed_to_stop, services_list)
    return _StopingDeferred


def suspend(service_name, *args, **kwargs):
    svc = services().get(service_name, None)
    if not svc:
        return None
    if svc.state == 'ON':
        return svc.suspend(*args, **kwargs)
    return None


def resume(service_name, *args, **kwargs):
    svc = services().get(service_name, None)
    if not svc:
        return None
    if svc.state == 'ON':
        return svc.resume(*args, **kwargs)
    return None


#------------------------------------------------------------------------------


def info(service_name):
    svc = services().get(service_name, None)
    if svc is None:
        service_name = 'service_' + service_name.replace('-', '_')
        svc = services().get(service_name, None)
    if svc is None:
        return None
    return svc.to_json()


#------------------------------------------------------------------------------


def restart(service_name, wait_timeout=None):
    global _StopingDeferred
    global _StartingDeferred
    restart_result = Deferred()

    def _on_started(start_result, stop_result, dependencies_results):
        if _Debug:
            lg.out(_DebugLevel, 'driver.restart._on_started : %s with %s, dependencies_results=%r' % (service_name, start_result, dependencies_results))
        try:
            stop_resp = {
                stop_result[0][1]: stop_result[0][0],
            }
        except:
            stop_resp = {
                'stopped': str(stop_result),
            }
        try:
            start_resp = {
                start_result[0][1]: start_result[0][0],
            }
        except:
            start_resp = {
                'started': str(start_result),
            }
        restart_result.callback([
            stop_resp,
            start_resp,
        ])
        return start_result

    def _on_failed(err):
        lg.warn('failed service %s in driver.restart() : %r' % (service_name, err))
        restart_result.errback(str(err))
        return None

    def _do_start(stop_result=None, dependencies_results=None):
        if _Debug:
            lg.out(_DebugLevel, 'driver.restart._do_start : %s' % service_name)
        start_defer = start(services_list=[service_name])
        start_defer.addCallback(_on_started, stop_result, dependencies_results)
        start_defer.addErrback(_on_failed)
        return start_defer

    def _on_stopped(stop_result, dependencies_results):
        if _Debug:
            lg.out(_DebugLevel, 'driver.restart._on_stopped : %s with %s' % (service_name, stop_result))
        _do_start(stop_result, dependencies_results)
        return stop_result

    def _do_stop(dependencies_results=None):
        if _Debug:
            lg.out(_DebugLevel, 'driver.restart._do_stop : %s' % service_name)
        stop_defer = stop(services_list=[service_name])
        stop_defer.addCallback(_on_stopped, dependencies_results)
        stop_defer.addErrback(_on_failed)
        return stop_defer

    def _on_timeout(err):
        if _Debug:
            lg.out(_DebugLevel, 'driver.restart._on_timeout : %s' % service_name)
        all_states = [_svc.state for _svc in services().values()]
        if 'INFLUENCE' in all_states or 'STARTING' in all_states or 'STOPPING' in all_states:
            restart_result.errback(Exception('timeout'))
            return err
        _do_stop()
        return None

    def _on_wait_timeout(err):
        if _Debug:
            lg.out(_DebugLevel, 'driver.restart._on_wait_timeout %s : %s' % (service_name, err.getErrorMessage()))
        # restart_result.errback(failure.Failure(Exception('timeout')))
        return None

    dl = []
    if _StopingDeferred and not _StopingDeferred.called:
        dl.append(_StopingDeferred)
    if _StartingDeferred and not _StartingDeferred.called:
        dl.append(_StartingDeferred)
    if wait_timeout:
        all_states = [_svc.state for _svc in services().values()]
        if 'INFLUENCE' in all_states or 'STARTING' in all_states or 'STOPPING' in all_states:
            wait_timeout_defer = Deferred()
            wait_timeout_defer.addErrback(_on_wait_timeout)
            wait_timeout_defer.addTimeout(wait_timeout, clock=reactor)
            dl.append(wait_timeout_defer)
    if not dl:
        dl.append(succeed(True))

    if _Debug:
        lg.args(_DebugLevel, service=service_name, wait_timeout=wait_timeout, starting=bool(_StartingDeferred), stoping=bool(_StopingDeferred))
    dependencies = DeferredList(dl, fireOnOneErrback=True, consumeErrors=True)
    dependencies.addCallback(_do_stop)
    dependencies.addErrback(_on_timeout)
    return restart_result


#------------------------------------------------------------------------------


def start_single(service_name):
    result = Deferred()
    _starting = Deferred()
    _stopping = Deferred()

    def _on_started(response):
        if response in ['started']:
            result.callback(True)
            return response
        if response in ['not_installed', 'failed', 'depends_off', 'stopped']:
            result.callback(False)
            return response
        raise Exception('bad response: %r' % response)

    def _on_stopped(response):
        if response != 'stopped' and response != 'depends_off':
            raise Exception('unpredicted response: %r' % response)
        svc.automat('start', _starting)
        return response

    def _on_failed(err, action):
        lg.warn('failed to %s service %s in driver.start_single()' % (action, service_name))
        return None

    _starting.addCallback(_on_started)
    _stopping.addCallback(_on_stopped)
    _starting.addErrback(_on_failed, 'start')
    _stopping.addErrback(_on_failed, 'stop')
    svc = services().get(service_name, None)
    if not svc:
        return succeed(False)
    if svc.state == 'ON':
        return succeed(True)
    if svc.state in ['STARTING']:
        svc.add_callback(_starting)
        return result
    if svc.state in ['STOPPING', 'INFLUENCE']:
        svc.add_callback(_stopping)
        return result
    svc.automat('start', _starting)
    return result


def stop_single(service_name):
    result = Deferred()
    _starting = Deferred()
    _stopping = Deferred()

    def _on_stopped(response):
        if response != 'stopped' and response != 'depends_off':
            raise Exception('unpredicted response: %r' % response)
        result.callback(True)
        return response

    def _on_started(response):
        if response in ['started']:
            svc.automat('stop', _stopping)
            return response
        if response in ['not_installed', 'failed', 'depends_off']:
            result.callback(True)
            return response
        raise Exception('bad response: %r' % response)

    def _on_failed(err, action):
        lg.warn('failed to %s service %s in driver.stop_single()' % (action, service_name))
        return None

    _starting.addCallback(_on_started)
    _stopping.addCallback(_on_stopped)
    _starting.addErrback(_on_failed, 'start')
    _stopping.addErrback(_on_failed, 'stop')
    svc = services().get(service_name, None)
    if not svc:
        return succeed(False)
    if svc.state == 'OFF':
        return succeed(True)
    if svc.state in ['STARTING']:
        svc.add_callback(_starting)
        return result
    if svc.state in ['STOPPING', 'INFLUENCE']:
        svc.add_callback(_stopping)
        return result
    svc.automat('stop', _stopping)
    return result


#------------------------------------------------------------------------------


def health_check(services_list=[]):
    if not services_list:
        services_list.extend(reversed(boot_up_order()))
    if _Debug:
        lg.out(_DebugLevel, 'driver.health_check with %d services' % len(services_list))
    dl = []
    for name in services_list:
        svc = services().get(name, None)
        if not svc:
            continue
        if not svc.enabled():
            continue
        service_health = svc.health_check()
        if isinstance(service_health, Deferred):
            dl.append(service_health)
        else:
            d = Deferred()
            d.callback(bool(service_health))
            dl.append(service_health)
    health_result = DeferredList(dl, consumeErrors=True)
    return health_result


#------------------------------------------------------------------------------


def get_network_configuration(services_list=[]):
    if not services_list:
        services_list.extend(reversed(boot_up_order()))
    if _Debug:
        lg.out(_DebugLevel, 'driver.get_network_info with %d services' % len(services_list))
    result = {
        'current_network': settings.CurrentNetworkName(),
    }
    for name in services_list:
        svc = services().get(name, None)
        if not svc:
            continue
        if not svc.enabled():
            continue
        network_configuration = svc.network_configuration()
        if network_configuration:
            result[name] = network_configuration
    return result


#------------------------------------------------------------------------------


def get_attached_dht_layers(services_list=[]):
    if not services_list:
        services_list.extend(reversed(boot_up_order()))
    if _Debug:
        lg.out(_DebugLevel, 'driver.get_attached_dht_layers with %d services' % len(services_list))
    result = {}
    for name in services_list:
        svc = services().get(name, None)
        if not svc:
            continue
        if not svc.enabled():
            continue
        dht_layers = svc.attached_dht_layers()
        if dht_layers:
            result[name] = dht_layers
    return result


#------------------------------------------------------------------------------


def populate_services():
    services_list = reversed(boot_up_order())
    for name in services_list:
        svc = services().get(name, None)
        if not svc:
            continue
        svc_data = svc.to_json()
        svc_data['event'] = None
        listeners.push_snapshot('service', snap_id=name, data=svc_data)


#------------------------------------------------------------------------------


def do_finish_starting():
    global _StartingDeferred
    if _Debug:
        lg.args(_DebugLevel, starting=bool(_StartingDeferred))
    _StartingDeferred = None


def do_finish_stoping():
    global _StopingDeferred
    if _Debug:
        lg.args(_DebugLevel, stoping=bool(_StopingDeferred))
    _StopingDeferred = None


#------------------------------------------------------------------------------


def on_service_callback(result, service_name):
    if _Debug:
        lg.out(_DebugLevel, 'driver.on_service_callback %s : [%s]' % (service_name, result))
    svc = services().get(service_name, None)
    if not svc:
        raise ServiceNotFound(service_name)
    if result == 'started':
        if _Debug:
            lg.out(_DebugLevel, '[%s] STARTED' % service_name)
        events.send('service-started', data=dict(name=service_name))
        relative_services = []
        for other_name in services().keys():
            if other_name == service_name:
                continue
            other_service = services().get(other_name, None)
            if not other_service:
                raise ServiceNotFound(other_name)
            if other_service.state == 'ON':
                continue
            for depend_name in other_service.dependent_on():
                if depend_name == service_name:
                    relative_services.append(other_service)
        if len(relative_services) > 0:
            for relative_service in relative_services:
                if not relative_service.enabled():
                    continue
                if relative_service.state == 'ON':
                    continue
                if _Debug:
                    lg.out(_DebugLevel, '    making attempt to start relative service %r' % relative_service)
                relative_service.automat('start')
    elif result == 'stopped':
        if _Debug:
            lg.out(_DebugLevel, '[%s] STOPPED' % service_name)
        events.send('service-stopped', data=dict(name=service_name))
        for depend_name in svc.dependent_on():
            depend_service = services().get(depend_name, None)
            if not depend_service:
                raise ServiceNotFound(depend_name)
            depend_service.automat('depend-service-stopped')
    elif result == 'depends_off':
        if _Debug:
            lg.out(_DebugLevel, '[%s] DEPENDS_OFF' % service_name)
        for depend_name in svc.dependent_on():
            depend_service = services().get(depend_name, None)
            if not depend_service:
                raise ServiceNotFound(depend_name)
            depend_service.automat('depend-service-stopped')
    svc_data = svc.to_json()
    svc_data['event'] = result
    listeners.push_snapshot('service', snap_id=service_name, data=svc_data)
    return result


def on_started_all_services(results):
    if _Debug:
        lg.out(_DebugLevel, 'driver.on_started_all_services results=%d' % len(results))
    reactor.callLater(0, do_finish_starting)  # @UndefinedVariable
    return results


def on_services_failed_to_start(err, services_list):
    if _Debug:
        lg.args(_DebugLevel, err=err, services_list=services_list)
    reactor.callLater(0, do_finish_starting)  # @UndefinedVariable
    return None


def on_stopped_all_services(results):
    if _Debug:
        lg.out(_DebugLevel, 'driver.on_stopped_all_services results=%d' % len(results))
    reactor.callLater(0, do_finish_stoping)  # @UndefinedVariable
    return results


def on_services_failed_to_stop(err, services_list):
    if _Debug:
        lg.args(_DebugLevel, err=err, services_list=services_list)
    reactor.callLater(0, do_finish_stoping)  # @UndefinedVariable
    return None


def on_service_enabled_disabled(path, newvalue, oldvalue, result):
    if not result:
        return
    if not path.endswith('/enabled'):
        return
    svc_name = path.replace('services/', 'service_').replace('/enabled', '').replace('-', '_')
    svc = services().get(svc_name, None)
    if svc:
        if newvalue == 'true':
            svc.automat('start')
        else:
            svc.automat('stop')
    else:
        lg.warn('%s not found: %s' % (svc_name, path))


#------------------------------------------------------------------------------


class ServiceAlreadyExist(Exception):
    pass


class RequireSubclass(Exception):
    pass


class ServiceNotFound(Exception):
    pass


class ServiceAlreadySuspended(Exception):
    pass


class ServiceWasNotSuspended(Exception):
    pass


#------------------------------------------------------------------------------


def main():
    from bitdust.main import settings
    lg.set_debug_level(20)
    settings.init()
    init()
    # print '\n'.join(_BootUpOrder)
    shutdown()
    settings.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# backup_tar.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (backup_tar.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: backup_tar.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import sys
from io import open

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet import threads
from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.append(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..'))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

#------------------------------------------------------------------------------

# Bytes Loop States:
BYTES_LOOP_EMPTY = 0
BYTES_LOOP_READY2READ = 1
BYTES_LOOP_CLOSED = 2

#------------------------------------------------------------------------------


class BytesLoop:
    def __init__(self, s=b''):
        self._buffer = s
        self._reader = None
        self._last_read = -1
        self._finished = False
        self._closed = False

    def read_defer(self, n=-1):
        if self._reader:
            raise Exception('already reading')
        self._reader = (
            Deferred(),
            n,
        )
        if len(self._buffer) > 0:
            chunk = self.read(n=n)
            d = self._reader[0]
            self._reader = None
            d.callback(chunk)
            return d
        if self._finished:
            chunk = b''
            d = self._reader[0]
            self._reader = None
            d.callback(chunk)
            return d
        return self._reader[0]

    def read(self, n=-1):
        before_bytes = len(self._buffer)
        chunk = self._buffer[:n]
        self._buffer = self._buffer[n:]
        after_bytes = len(self._buffer)
        self._last_read = len(chunk)
        if _Debug:
            lg.args(_DebugLevel, before_bytes=before_bytes, after_bytes=after_bytes, chunk_bytes=len(chunk))
        return chunk

    def write(self, chunk):
        reactor.callFromThread(self._write, chunk)  # @UndefinedVariable

    def _write(self, chunk):
        self._buffer += chunk
        if _Debug:
            lg.args(_DebugLevel, buffer_bytes=len(self._buffer), chunk_bytes=len(chunk))
        if len(self._buffer) > 0:
            if self._reader:
                chunk = self.read(n=self._reader[1])
                d = self._reader[0]
                self._reader = None
                d.callback(chunk)

    def close(self):
        if self._reader:
            d = self._reader[0]
            self._reader = None
            d.callback(b'')
        self._closed = True
        self._buffer = b''

    def kill(self):
        self.close()

    def mark_finished(self):
        self._finished = True

    def state(self):
        if self._closed:
            return BYTES_LOOP_CLOSED
        if len(self._buffer) > 0:
            if self._reader:
                return BYTES_LOOP_EMPTY
            return BYTES_LOOP_READY2READ
        if self._last_read > 0:
            return BYTES_LOOP_READY2READ
        if self._finished:
            return BYTES_LOOP_EMPTY
        if not self._reader:
            return BYTES_LOOP_READY2READ
        return BYTES_LOOP_EMPTY


#------------------------------------------------------------------------------


def backuptarfile_thread(filepath, arcname=None, compress=None):
    """
    Makes tar archive of a folder inside a thread.
    Returns `BytesLoop` object instance which can be used to read produced data in parallel.
    """
    if not os.path.isfile(filepath):
        lg.err('file %s not found' % filepath)
        return None
    if arcname is None:
        arcname = os.path.basename(filepath)
    p = BytesLoop()

    def _run():
        from bitdust.storage import tar_file
        ret = tar_file.writetar(
            sourcepath=filepath,
            arcname=arcname,
            subdirs=False,
            compression=compress or 'none',
            encoding='utf-8',
            fileobj=p,
        )
        p.mark_finished()
        if _Debug:
            lg.out(_DebugLevel, 'backup_tar.backuptarfile_thread writetar() finished')
        return ret

    reactor.callInThread(_run)  # @UndefinedVariable
    return p


def backuptardir_thread(directorypath, arcname=None, recursive_subfolders=True, compress='bz2'):
    """
    Makes tar archive of a single file inside a thread.
    Returns `BytesLoop` object instance which can be used to read produced data in parallel.
    """
    if not bpio.pathIsDir(directorypath):
        lg.err('folder %s not found' % directorypath)
        return None
    if arcname is None:
        arcname = os.path.basename(directorypath)
    p = BytesLoop()

    def _run():
        from bitdust.storage import tar_file
        ret = tar_file.writetar(
            sourcepath=directorypath,
            arcname=arcname,
            subdirs=recursive_subfolders,
            compression=compress or 'none',
            encoding='utf-8',
            fileobj=p,
        )
        p.mark_finished()
        if _Debug:
            lg.out(_DebugLevel, 'backup_tar.backuptardir_thread writetar() finished')
        return ret

    reactor.callInThread(_run)  # @UndefinedVariable
    return p


def extracttar_thread(tarfile, outdir, mode='r:bz2'):
    """
    Opposite method, extract files and folders from ".tar" file inside a thread.
    """
    if not os.path.isfile(tarfile):
        lg.err('path %s not found' % tarfile)
        return None
    if _Debug:
        lg.out(_DebugLevel, 'backup_tar.extracttar_thread tarfile=%s' % tarfile)

    def _run():
        if _Debug:
            lg.out(_DebugLevel, 'backup_tar.extracttar_thread._run outdir=%s' % outdir)
        from bitdust.storage import tar_file
        ret = tar_file.readtar(
            archivepath=tarfile,
            outputdir=outdir,
            encoding='utf-8',
            mode=mode,
        )
        return ret

    return threads.deferToThread(_run)  # @UndefinedVariable


#------------------------------------------------------------------------------


def test_in_thread():
    fout = open('out.tar', 'wb')

    def _read(p):
        if p.state() == BYTES_LOOP_CLOSED:
            print('closed')
            fout.close()
            reactor.stop()  # @UndefinedVariable
            return
        if p.state() == BYTES_LOOP_EMPTY:
            print('empty')
            reactor.callLater(0, _read, p)  # @UndefinedVariable
            return
        chunk = p.read()
        fout.write(chunk)
        print(len(chunk))
        if not chunk:
            print('empty chunk')
            fout.close()
            reactor.stop()  # @UndefinedVariable
            return
        reactor.callLater(0, _read, p)  # @UndefinedVariable

    def _go():
        p = backuptarfile_thread(sys.argv[1], arcname='asd')
        _read(p)

    reactor.callLater(0, _go)  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable


if __name__ == '__main__':
    test_in_thread()



#!/usr/bin/python
# config_details.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (config_details.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: config_details
"""


def raw():
    """
    In case option does not have any description here it will be marked as "internal".
    The "internal" options are not displayed in the UI, but still can be manipulated via API or command line.
    """
    return """
{interface/api/auth-secret-enabled}

{interface/api/rest-http-enabled}

{interface/api/rest-http-port}

{interface/api/web-socket-enabled}

{interface/api/web-socket-port}

{interface/ftp/enabled}

{interface/ftp/port}

{logs/api-enabled} log API calls
Enable logging of all API calls that reach the main process to the `~/.bitdust/logs/api.log` file.

{logs/automat-events-enabled} log state machines events
This option enables logging of all events that are submitted to state machines, see `~/.bitdust/logs/automats.log` file.

{logs/automat-transitions-enabled} log state machines transitions
By enabling this option, you can also monitor the life cycle of each state machine running in the main process, see `~/.bitdust/logs/automats.log` file.

{logs/debug-level} debug level
Higher values of `debug-level` option will produce more log messages in the console output, see `~/.bitdust/logs/stdout.log` file.

{logs/packet-enabled} log network packets
This option enables logging of all incoming & outgoing peer-to-peer packets - very helpful when analyzing/debugging network communications, see `~/.bitdust/logs/packet.log` file.

{logs/stream-enabled}

{logs/stream-port}

{logs/traffic-enabled}

{logs/traffic-port}


{paths/backups}

{paths/customers}

{paths/messages}

{paths/receipts}

{paths/restore}


[personal/private-key-size] RSA private key size
This value you specified when created your identity file for the first time.

{services/accountant/enabled} enable accountant service
The service is under development.

{services/backup-db/enabled} enable backup-db service
The `backup-db` service is responsible for storing service information about your files uploaded to the network, such as: file name, version, size.

{services/backups/enabled} enable data uploading
This service enables uploading of encrypted personal data to the BitDust network.

{services/backups/block-size} preferred block size
Large files are also divided by chunks during uploading, that parameter defines preferred block size in bytes
which is used to split the raw data during uploading.

{services/backups/max-block-size} maximum block size
Maximum block size in bytes which used to split the raw data during uploading process.
The actual block size is calculated depending on size of the particular backup to optimize performance and data storage.
If you plan to do a large file uploads - set higher values to increase the performance.

{services/backups/keep-local-copies-enabled} keep locally copies of uploaded files
Enable this to keep a copy of every uploaded file on your local disk as well as on remote machines of your suppliers.
This increases data reliability, rebuilding performance and decrease network load, but consumes much storage space of your own device.

{services/backups/max-copies} number of copies
When you upload same file multiple times data is not overwritten - another version is created instead.
This value indicates how many versions of same uploaded file must be stored on remote suppliers.
The oldest copies are removed automatically. A value of `0` indicates unlimited number of versions.

{services/backups/wait-suppliers-enabled} extra check after 24 hours
When critical amount of your suppliers become unreliable - your uploaded data is lost completely.
Enable this option to wait for 24 hours after any file upload and perform an extra check of all suppliers before cleaning up the local copy.
This is a compromise solution that does not sacrifice reliability but also decrease local storage consumption.

{services/broadcasting/enabled} send & receive encrypted broadcast messages
The service is under development.

{services/broadcasting/routing-enabled}

{services/broadcasting/max-broadcast-connections}

{services/contract-chain/enabled} accounting of spent & donated resources
The service is under development.

{services/customer/enabled} store encrypted data on other nodes
To be able to upload files to the BitDust network other nodes, called suppliers, must be already connected and allocate storage space for you.
The `customer` service is responsible for maintaining real-time connections with all of your suppliers and requesting storage space automatically.

{services/customer/needed-space} needed space
How much storage space do you require to store your uploaded files?
**WARNING!** When your uploaded files exceed the quota size you set here the older versions will be automatically erased by suppliers.

{services/customer/suppliers-number} number of suppliers
Number of suppliers who stores your encrypted files. You may select one of the following values: 2, 4, 7, 13, 18, 26, 64.
**WARNING!** You will lose all of your uploaded files when changing preferred number of suppliers.

{services/customer-contracts/enabled} digitally signed customer contracts
The service is under development.

{services/customer-family/enabled} maintain service info for my customers
When `supplier` service is enabled and started your device will also write additional service information to distributed hash table.
This way people are able to access encrypted shared files of each other on different suppliers across the network.

{services/customer-patrol/enabled} reject customers who are out of quota
When you are running `supplier` service other users are able to store files on your device.
The `customer-patrol` service makes sure that your customers do not exceed requested and agreed storage quotas.

{services/customer-patrol/customer-idle-days} customer idle limit
This option enables few periodic checks and will auto-reject customers who are no longer using your device to store any data.
Specify number of days of customer inactivity before user will be rejected and data erased automatically.
A value of `0` disables the feature.

{services/customer-support/enabled} assist my customers
This service performs periodic synchronization for your customers when you also enabled the `supplier` service and thus agreed to store encrypted data for them.
Your device will try to reconnect with all known customers and keep them up to date with the full list of files they own.

{services/data-disintegration/enabled} enable RAID & ECC data processing
Before the encrypted data is sent to the target storage nodes, it will be chunked and processed.
Each fragment corresponds to a supplier position in accordance with the number of suppliers you have selected.
Data processing includes error correction codes and redundancy to enable restoration of missing fragments in the future.

{services/data-motion/enabled} streaming of encrypted data
The service creates a queue of incoming and outgoing encrypted fragments of your data when uploading and downloading from the nodes of remote providers.

{services/data-motion/supplier-request-queue-size} concurrent outgoing packets
Determines the maximum number of encrypted data packets sent at the same time.
Affects the speed of data uploading to the suppliers' machines.

{services/data-motion/supplier-sending-queue-size} concurrent incoming packets
Determines the maximum number of simultaneously received encrypted data packets.
Affects the speed of data downloading from the suppliers' machines.

{services/entangled-dht/enabled} distributed hash-table node
Your device becomes one of the peers in the DHT network.
Provides the ability to read and write to a distributed hash table and access networking service layers.

{services/entangled-dht/udp-port} UDP port number
Set a UDP port number to be used for distributed hash-table communications.

{services/entangled-dht/known-nodes} override seed-nodes list
For the initial entrance to the DHT network, a list of so-called seed-nodes is required.
In the future, your device will work autonomously - thanks to the local routing table.
This value allows you to override the initial set of hard-coded seed-nodes and is intended for advanced software use.

{services/entangled-dht/attached-layers} override attached DHT layers
The software supports several local routing tables for the ability to store service data on the DHT at different parallel "layers".
This allows the network services to be separated from each other and also makes possible to run lookups in the DHT more faster.
On startup, your device will be automatically connected to some of the layers.
This value overrides this list and is intended for advanced software use.

{services/employer/enabled} search & connect with available suppliers
In order to store data on the network, you must already have your suppliers ready and accepting your uploads.
The `employer` network service automatically searches for new suppliers through the DHT network and monitors their reliability.

{services/employer/replace-critically-offline-enabled} automatically replace unreliable suppliers
The moment a critical number of your suppliers become unreliable, your uploaded data is no longer available for downloading.
When this option is enabled, your software will automatically replace the unreliable supplier when needed and reconstruct missing fragments.

{services/employer/candidates} candidates list
Place here a comma-separated list of nodes identifiers, in IDURL format, to be used first when the service tries to find a new supplier for you.
This way you can control who will be your supplier and where your data is stored.
Option is intended for advanced software use.

{services/gateway/enabled} enable encrypted peer-to-peer traffic
You can use `TCP`, `UDP`, and other network protocols to communicate with people on the network.
The `gateway` service controls application transport protocols and all encrypted packets passing through and reaching application engine.
Every incoming packet is digitally verified here, processed and sent to the underlying services.

{services/gateway/p2p-timeout} peer-to-peer reply timeout
Due to network failures or slowness, signed peer-to-peer packets are considered as "undelivered" without receiving a confirmation of delivery within the specified number of seconds.

{services/http-connections/enabled} HTTP enabled
This will allow BitDust to use the HTTP protocol for service data and encrypted traffic

{services/http-connections/http-port} HTTP port number
That option defines a port number to be used to listen for incoming HTTP connections.

{services/http-transport/enabled} use HTTP as a transport protocol
Your BitDust node must have at least one active and reliable transport protocol to be able
to connect to other nodes via Internet.
The service is under development.

{services/http-transport/receiving-enabled} receiving via HTTP
Disable this option if you do not want other nodes to send you encrypted data over HTTP and use transport protocol in unidirectional mode.

{services/http-transport/sending-enabled} sending via HTTP
Disable this option if you do not want to send encrypted data over HTTP and use transport protocol in unidirectional mode.

{services/http-transport/priority} priority
You can change this setting if you want BitDust to use the `http-transport` more often than other transport protocols.
Lower values have higher priority.

{services/identity-propagate/enabled} signed identity files exchange
Your identity file stores service information that is used by other nodes when they connect to your device to transfer data.
This service manages the synchronization and propagating of your identity file, and also remembers the identity files of other users.

{services/identity-propagate/known-servers}

{services/identity-propagate/preferred-servers}

{services/identity-propagate/min-servers}

{services/identity-propagate/max-servers}

{services/identity-propagate/automatic-rotate-enabled} automatic identity rotation
Automatically replaces untrusted ID servers - nodes that store a copy of your identity file.
This way, your device will remain identifiable on the network even when it is offline.

{services/identity-propagate/health-check-interval-seconds} health check interval
The period in seconds between checks that could potentially cause an automatic identity rotation when your primary ID server is down.

{services/identity-server/enabled} start Identity server
You can start own Identity server and store identity files of other users on your machine.
This will help others to join and operate in the BitDust network.

{services/identity-server/host} your server host name
Each ID server must have a unique address within the network, set your external IP address or domain name here.

{services/identity-server/tcp-port} TCP port number
TCP transport protocol is used for receiving digitally-signed identity files from other nodes.
That option defines TCP port number to be used to listen for incoming connections.

{services/identity-server/web-port} HTTP port number
Port number for Identity server HTTP interface.
This is used when other nodes are "reading" identity files from your Identity server.

{services/ip-port-responder/enabled} reply STUN requests
Helps other hosts connect to the network.
When the device is behind a NAT router, additional steps are required to determine the real external IP address.

{services/keys-registry/enabled} enables keys management
Keeps track of all your RSA keys that are used to encrypt files and messages.

{services/keys-storage/enabled} enable keys synchronization
The service maintains an encrypted backup of every RSA key you possess on your suppliers so that you can recover the keys and encrypted data in case of loss.

{services/keys-storage/reset-unreliable-backup-copies} reset unreliable backup copies
In worst scenario, when your suppliers are not reliable anymore and uploaded data got lost, software will be blocked trying to synchronize your keys.
This setting helps to overcome this situation by using locally stored copies of the keys.

{services/list-files/enabled} exchange service data with suppliers
The supplier does not know the real name of the file you uploaded to the network, but stores indexed fragments of your encrypted data.
The `list-files` service synchronizes index data and file lists with your suppliers.

{services/message-broker/enabled} help others to stream encrypted messages
The `message-broker` service stores and organizes a message queue of encrypted JSON message thru your device.
Other users are able to connect to your device to exchange messages and thus you will supports group chats for other users particularly.

{services/message-broker/archive-chunk-size} archive chunk size
Old messages in the queue are split into blocks, encrypted and archived on the suppliers nodes.

{services/message-broker/message-ack-timeout} message acknowledgment timeout
Due to network failures or slowness, peer-to-peer messages are considered as "undelivered" without receiving a confirmation of delivery within the specified number of seconds.

{services/message-broker/broker-negotiate-ack-timeout} brokers synchronization timeout
Each queue requires three message brokers at all time (main, secondary, third) and those three must cooperate with each other periodically to stay in sync.
This value defines a timeout for an acknowledgment from another broker.
When one of the brokers broker is unreliable, eventually all three will be "rotated" and "dead" broker will be replaced by fresh one.

{services/message-broker/preferred-brokers}

{services/message-history/enabled} store messaging history
All your conversations are stored in a local SQLite3 database and are quickly searchable.

{services/miner/enabled} enable blockchain mining
The service is under development.

{services/my-data/enabled} remote files indexing
The service synchronizes with my suppliers in real time and prepares the encrypted storage environment for other dependent network services.

{services/my-ip-port/enabled} detect my external IP via other nodes
The service relies on other nodes and uses the STUN protocol to recognize your public IP address when connecting to the network.

{services/network/enabled} network is enabled
Basic network service of the application. If you disable it, all other network services will be turned off as well and your device will go offline.

{services/network/proxy/enabled}

{services/network/proxy/host}

{services/network/proxy/password}

{services/network/proxy/port}

{services/network/proxy/ssl}

{services/network/proxy/username}

{services/network/receive-limit} limit incoming traffic
The value in bytes per second to limit incoming network traffic.
Set value to `0` if you do not want to limit your incoming traffic.

{services/network/send-limit} limit outgoing traffic
The value in bytes per second to limit outgoing network traffic.
Set value to `0` if you do not want to limit your outgoing traffic.

{services/nodes-lookup/enabled} lookup nodes via DHT
The `network-lookup` service allows nodes to find each other in distributed hash table and initiate direct peer-to-peer connection when resources need to be allocated.

{services/p2p-hookups/enabled} peer-to-peer connections
The service is responsible for starting and maintaining direct network connections with other nodes in order to be able to transfer personal data instantly.

{services/p2p-notifications/enabled} events & notifications delivery
Supports other network services while sending and receiving service events.

{services/personal-messages/enabled} personal messages
Allows you to read incoming encrypted messages even when your device is disconnected.
Other nodes, called "message brokers", will maintain a stream of messages that you can subscribe to and receive at any time.

{services/private-groups/enabled} group communications
Allows the exchange of encrypted messages with multiple users using the same secret key through message brokers.

{services/private-groups/message-ack-timeout} acknowledgment timeout
Defines a timeout in seconds for an acknowledgment after sending encrypted message to message broker.

{services/private-groups/broker-connect-timeout} connect timeout
Defines a timeout in seconds for an acknowledgment from message broker while connecting to the group.

{services/private-groups/preferred-brokers}

{services/private-messages/enabled} encrypted peer-to-peer messaging
Enables peer-to-peer sending & receiving of encrypted JSON messages and also supports underlying services.

{services/private-messages/acknowledge-unread-messages-enabled}

{services/proxy-server/enabled} help others transmit encrypted traffic
Your device will act as an intermediate node and help other users stay connected to the network and exchange encrypted packets with each other.
This will consume your Internet traffic.

{services/proxy-server/routes-limit} maximum simultaneous connections
This option sets a limit on the number of connections your device will support at any given time.

{services/proxy-server/current-routes}

{services/proxy-transport/enabled} use intermediate nodes
The `proxy-transport` network protocol allows you to use another user's device as an intermediate node for encrypted data exchange.
This solves certain problems when connecting to the network and also makes your network connection more anonymous.

{services/proxy-transport/receiving-enabled} receiving enabled
Disable this option if you do not want other nodes to send you encrypted data via intermediate node and use that transport protocol in unidirectional mode.

{services/proxy-transport/sending-enabled} sending enabled
Disable this option if you do not want to send encrypted data via intermediate node and use that transport protocol in unidirectional mode.

{services/proxy-transport/priority} priority
You can change this setting if you want BitDust to use the `proxy-transport` more often than other transport protocols.
Lower values have higher priority.

{services/proxy-transport/my-original-identity}

{services/proxy-transport/current-router}

{services/proxy-transport/preferred-routers}

{services/proxy-transport/router-lifetime-seconds}

{services/rebuilding/enabled} automatic data rebuilding
When one of your suppliers is not available for some reason, all encrypted fragments that he stored for you are lost.
The `rebuilding` service will automatically download the available fragments from those suppliers that are still online, and "rebuild" the lost fragments that the new supplier receives.
**WARNING!** At the moment when a critical number of fragments are lost, downloading data is no longer possible.

{services/restores/enabled} enable data downloading
Controls network connections and incoming data streams when downloading encrypted fragments from suppliers nodes.

{services/shared-data/enabled} enable data sharing
Makes possible decentralized sharing of encrypted files with other users.

{services/supplier/enabled} donate own disk space to other users
To make it possible to store any files on the network, some other nodes must already be connected and provide storage space.
The `supplier` service allocates a part of your disk and serves encrypted files uploaded to your device from remote nodes.

{services/supplier/donated-space} donated space
The amount of storage space you want to donate to other users.

{services/supplier-contracts/enabled} digitally signed supplier contracts
The service is under development.

{services/tcp-connections/enabled} TCP enabled
This will allow BitDust to use the TCP protocol for service data and encrypted traffic

{services/tcp-connections/tcp-port}
Port number to listen for incoming TCP connections.

[services/tcp-connections/upnp-enabled]
The feature is under development and temporarily disabled.

{services/tcp-transport/enabled} use TCP as a transport protocol
BitDust can use different protocols to transfer packets across the network.
Your node must have at least one active and reliable transport protocol to be able
to connect to other nodes via Internet.

{services/tcp-transport/receiving-enabled} receiving via TCP
Disable this option if you do not want other nodes to send you encrypted data over TCP and use `tcp-transport` in unidirectional mode.

{services/tcp-transport/sending-enabled} sending via TCP
Disable this option if you do not want to send encrypted data over TCP and use `tcp-transport` in unidirectional mode.

{services/tcp-transport/priority} priority
You can change this setting if you want BitDust to use the `tcp-transport` more often than other transport protocols.
Lower values have higher priority.

{services/udp-datagrams/enabled} UDP enabled
This will allow BitDust to use the UDP protocol for service data and encrypted traffic.

{services/udp-datagrams/udp-port}
Port number to listen for incoming UDP datagrams.

[services/udp-transport/enabled] use UDP as a transport protocol
BitDust can use different protocols to transfer packets across the network.
Your node must have at least one active and reliable transport protocol to be able
to connect to other nodes via Internet.
The service is under development.

[services/udp-transport/receiving-enabled]
Disable this option if you do not want other nodes to send you encrypted data over UDP and use `udp-transport` in unidirectional mode.

[services/udp-transport/sending-enabled]
Disable this option if you do not want to send encrypted data over UDP and use `udp-transport` in unidirectional mode.

[services/udp-transport/priority]
You can change this setting if you want BitDust to use the `udp-transport` more often than other transport protocols.
Lower values have higher priority.
"""



import log


def aliases_update(node, db_handler_instance):

    db_handler_instance.index_cursor.execute('SELECT block_height FROM aliases ORDER BY block_height DESC LIMIT 1;')
    try:
        alias_last_block = int(db_handler_instance.index_cursor.fetchone()[0])
    except:
        alias_last_block = 0

    node.logger.app_log.warning('Alias anchor block: {}'.format(alias_last_block))

    db_handler_instance.h.execute('SELECT block_height, address, openfield FROM transactions WHERE operation = ? AND block_height >= ? AND reward = 0 ORDER BY block_height ASC, timestamp ASC;', (
        'alias:register',
        alias_last_block,
    ))
    #include the anchor block in case indexation stopped there
    result = db_handler_instance.h.fetchall()

    for openfield in result:
        node.logger.app_log.warning(f'Processing alias registration: {openfield[2]}')
        try:
            db_handler_instance.index_cursor.execute('SELECT * from aliases WHERE alias = ?', (openfield[2], ))
            dummy = db_handler_instance.index_cursor.fetchall()[0]  #check for uniqueness
            node.logger.app_log.warning(f'Alias already registered: {openfield[2]}')
        except:
            db_handler_instance.index_cursor.execute('INSERT INTO aliases VALUES (?,?,?)', (openfield[0], openfield[1], openfield[2]))
            db_handler_instance.index.commit()
            node.logger.app_log.warning(f'Added alias to the database: {openfield[2]} from block {openfield[0]}')


if __name__ == '__main__':
    app_log = log.log('aliases.log', 'WARNING', True)
